

from baopig import *


# --- DEBUG ---
class DebugZone(Zone, Closable):

    def __init__(self, scene):

        Zone.__init__(
            self,
            parent=scene,
            size=scene.size,
            background_color=(0, 0, 0, 0),
            layer=scene.debug_layer
        )

        self.set_style_for(Text, color="black")

        self._pointed = None
        self.highlighter = None

        h = 200
        self.debug_zone = Zone(
            parent=self,
            size=(self.w - 10, h),
            background_color=(255, 255, 255, 145),
            name=self.name + " -> debug_zone",
            pos=(5, -5), pos_location="bottomleft", pos_ref_location="bottomleft"
        )
        # TODO : self.debug_zone.style.font.file = None

        presentators_zone = Zone(
            parent=self.debug_zone,
            pos=(5, 5),
            size=(80, self.debug_zone.h),
            name=self.debug_zone.name + " -> presentators_zone"
        )
        trackers_zone = Zone(
            parent=self.debug_zone,
            pos=(presentators_zone.right, 5),
            size=(4000, self.debug_zone.h),  # TODO : how to say 'I define the height but not the width ?' or 'width adapts to the cell'
            name=self.debug_zone.name + " -> trackers_zone"
        )

        if True:
            # FPS TRACKER
            fps_presentator = Text(
                parent=presentators_zone,
                text="FPS : ",
                pos=(0, 5),
                name="FPS_presentator")
            DynamicText(
                parent=trackers_zone,
                get_text=lambda: self.parent.painter.get_current_fps(),
                pos=(0, fps_presentator.top),
                name="FPS_tracker")

            # MOUSE TRACKER
            mouse_pos_presentator = Text(
                parent=presentators_zone,
                text="Mouse : ",
                pos=(0, fps_presentator.bottom),
                name="mouse_pos_presentator")
            DynamicText(
                parent=trackers_zone,
                get_text=PrefilledFunction(lambda: mouse.pos),
                pos=(0, mouse_pos_presentator.top),
                name="mouse_pos_tracker")

            pointed_comp_presentator = Text(
                parent=presentators_zone,
                text="Hovered : ",
                pos=(0, mouse_pos_presentator.bottom),
                name="pointed_comp_presentator")

            # NAME TRACKER
            name_presentator = Text(
                parent=presentators_zone,
                text="- name : ",
                pos=(0, pointed_comp_presentator.bottom),
                name="name_presentator")
            DynamicText(
                parent=trackers_zone,
                get_text=lambda: mouse.pointed_comp.name if mouse.pointed_comp else None,
                pos=(0, name_presentator.top),
                name="name_tracker")

            # HITBOX TRACKER
            hitbox_presentator = Text(
                parent=presentators_zone,
                text="- hitbox : ",
                pos=(0, name_presentator.bottom),
                name="hitbox_presentator")
            DynamicText(
                parent=trackers_zone,
                get_text=lambda: mouse.pointed_comp.hitbox if mouse.pointed_comp else None,
                pos=(0, hitbox_presentator.top),
                name="hitbox_tracker")

            # PARENT TRACKER
            parent_presentator = Text(
                parent=presentators_zone,
                text="- parent : ",
                pos=(0, hitbox_presentator.bottom),
                name="parent_presentator")
            DynamicText(
                parent=trackers_zone,
                get_text=lambda: mouse.pointed_comp.parent if mouse.pointed_comp else None,
                pos=(0, parent_presentator.top),
                name="parent_tracker")

            # POINTED TRACKER
            # pointed_presentator = Text(
            #     parent=presentators_zone,
            #     text="Pointed : ",
            #     pos=(0, parent_presentator.bottom),
            #     name="pointed_presentator")
            # DynamicText(
            #     parent=trackers_zone,
            #     get_text=lambda: repr(mouse._get_pointed_comp()),
            #     pos=(0, pointed_presentator.top),
            #     name="pointed_tracker")

            # TODO : mouse.pointed_comp tracker

        self.print_text = Text(
            parent=self.debug_zone,
            text="This text is aimed to debug",
            pos=(5, self.debug_zone.h - 5 - 15),
            name=self.debug_zone.name + " -> print_text"
        )

        # TODO : grid pack with margin

        def print_pointed_comp(*args):
            if hasattr(mouse.pointed_comp.parent, "text"):
                self.print(mouse.pointed_comp.parent.text)
            # self.print(repr(mouse.pointed_comp))
        # self.handle_keydown[keyboard.SPACE].add(print_pointed_comp)

        self.connect("update", self.parent.signal.RESIZE)
        self.connect("update_pointed_outline", mouse.signal.MOTION)
        self.connect("update_pointed_outline", mouse.signal.DRAG)

        self.update_pointed_outline()
        self.debug_zone.adapt(self.debug_zone.default_layer, horizontally=False)
        self.set_nontouchable()
        self.set_always_dirty()

    is_debugging = property(lambda self: not self.is_sleeping)

    def asleep(self):

        self.highlighter.hide()
        super().asleep()

    def close(self):

        self.kill()

    def print(self, obj):

        self.print_text.set_text(str(obj))
        self.debug_zone.adapt(self.debug_zone.default_layer, horizontally=False)

    def start_debugging(self):

        if self.is_debugging:
            return

        self.wake()

    def stop_debugging(self):

        if not self.is_debugging:
            return

        self.asleep()

    def toggle_debugging(self):

        if self.is_debugging:
            self.stop_debugging()
        else:
            self.start_debugging()

    def update(self, old_size):

        self.resize(*self.parent.size)
        self.debug_zone.resize_width(self.size[0] - 10)
        self.update_pointed_outline()

    def update_pointed_outline(self, *args):

        pointed = mouse.pointed_comp
        if pointed and self.is_awake:
            if self._pointed == pointed: return
            self._pointed = pointed
            if self.highlighter is None:
                self.highlighter = Highlighter(
                    parent=self,
                    target=self._pointed,
                    color=(0, 255, 0),
                    width=1,
                    name=self.name + ".highlighter"
                )
                self.highlighter.move_behind(self.debug_zone)
            else:
                self.highlighter.config(target=self._pointed)
        else:
            self._pointed = None
            if self.highlighter is not None:
                self.highlighter.hide()

    def wake(self):

        self.highlighter.show()
        super().wake()
        self.update_pointed_outline()






import threading
import pygame
from baopig.pybao.objectutilities import TypedList
from .widget import debug_with_assert
from .layer import Layer


# TODO : many widgets in one cell ?


class Cell:

    def __init__(self, grid, row_index, col_index):

        self._grid = grid
        self._col_index = col_index
        self._row_index = row_index

    col = property(lambda self: self._grid.get_col(self._col_index))
    row = property(lambda self: self._grid.get_row(self._row_index))

    def get_data(self):
        """
        Return a cell's content
        None, a Widget or a list of components (if implemented, but not yet)
        """
        with self._grid._lock:
            return self._grid._data[self._row_index][self._col_index]

    def get_rect(self):
        """
        Return the cell's position and size inside a rectstyle object
        """
        return pygame.Rect(self.col.left, self.row.top, self.col.width, self.row.height)


class Column:
    """
    A Column can be seen as a list of cells with the same 'col' value.
    It is the manager of GridLayer.cols and can read GridLayer.data
    """

    def __init__(self, grid, index):

        assert isinstance(grid, GridLayer)
        assert isinstance(index, int) and index >= 0
        if debug_with_assert: assert index == len(grid._cols)  # maybe not, if col inserted, but not implemented yet

        self._grid = grid
        self._grid_layer = grid._layer
        self._grid_data = grid._data  # shared ressource
        self._col_index = index
        self._width = grid.col_width
        if self.is_first: self._left = 0
        else: self._left = self.get_previous_col().right  # TODO : + margin

        grid._cols.append(self)  # TODO : insert ?

    def __getitem__(self, row):
        """
        Return the value of the cell in this col, indexed by 'col'
        """
        return self._grid_data[row][self._col_index]

    def __iter__(self):

        for data_row in self._grid_data:
            yield data_row[self._col_index]

    def __contains__(self, item):

        for data_row in self._grid_data:
            if data_row[self._col_index] is item:
                return True
        return False

    components = property(lambda self: tuple(comp for comp in self if comp is not None))
    icomponents = property(lambda self: (comp for comp in self if comp is not None))
    is_adaptable = property(lambda self: self._width is None)
    is_first = property(lambda self: self._col_index is 0)
    is_last = property(lambda self: self._col_index is len(self._grid.cols) - 1)
    left = property(lambda self: self._left)
    right = property(lambda self: self.left + self.width)

    def _update_width(self):
        """
        Adapt the col width to the required width
        """
        w = self.width
        if not self.is_last and self.get_next_col().left == self.left + w:
            return
        if debug_with_assert:
            if self.components:
                has_window = self.components[0].window is not None
            for comp in self.icomponents:
                assert has_window == (comp.window is not None)

        left = self.left
        for comp in self.icomponents:
            row = self._grid.get_row(comp.row)
            comp.set_window((left, row.top, w, row.height), follow_movements=True)
            if comp.sticky is not None:
                comp.origin.unlock()
                comp.origin.config(pos=getattr(self.get_cell(comp.row).get_rect(), comp.sticky))
                comp.origin.lock()
            if debug_with_assert: assert comp.window == self.get_cell_rect(comp.row)

        if not self.is_last:
            self.get_next_col()._update_left(left + w)

    def _update_left(self, left):

        dx = left - self.left
        if dx is 0: return
        self._left = left
        for comp in self.icomponents:
            comp.origin.unlock()
            comp.move(dx=dx)
            comp.origin.lock()

        if not self.is_last:
            self.get_next_col()._update_left(left + self.width)

    def is_empty(self):

        for cell in self:
            if cell is not None:
                return False
        return True

    def get_cell(self, row):
        """
        Return the cell whose column is indexed by 'col'
        """
        return Cell(self._grid, row, self._col_index)

    def get_cell_rect(self, row):
        """
        Optimized shortcut instead of col.get_cell(...).get_rect()
        """
        return self.left, self._grid.get_row(row).top, self.width, self._grid.get_row(row).height

    def get_next_col(self):

        if self.is_last: return
        return self._grid.get_col(self._col_index + 1)

    def get_previous_col(self):
        """
        Return the col above self
        """
        if self.is_first: return None
        return self._grid.get_col(self._col_index - 1)

    def get_rect(self):
        """
        Return the col's position (relative to the container) and size inside a rectstyle object
        """
        return self.left, self._grid_layer.container.top, self.width, self._grid_layer.container.h

    def get_width(self):

        if self._width is not None:
            return self._width

        if self.is_empty():
            return 0

        return max(comp.rect.w for comp in self if comp is not None)

    width = property(get_width)

    def set_width(self, width):
        """
        Set the col's width
        If width is None, the col will adapt to its widest component
        """
        assert (width is None) or isinstance(width, (int, float)) and width >= 0
        if not self._grid.cols_are_adaptable:
            raise PermissionError("This grid has a fixed col_width : {}".format(self._grid.col_width))
        self._width = width
        self._update_width()


class Row:
    """
    A Row can be seen as a list of cells with the same 'row' value.
    It is the manager of GridLayer.rows and can read GridLayer.data
    """
    def __init__(self, grid, index):

        assert isinstance(grid, GridLayer)
        assert isinstance(index, int) and index >= 0
        if debug_with_assert: assert index == len(grid._rows)  # maybe not, if row inserted, but not implemented yet

        self._grid = grid
        self._grid_layer = grid._layer
        self._grid_data = grid._data  # shared ressource
        self._row_index = index
        self._height = grid.row_height
        if self.is_first: self._top = 0
        else: self._top = self.get_previous_row().bottom

        grid._rows.append(self)

    def __getitem__(self, col):
        """
        Return the value of the cell in this row, indexed by 'col'
        """
        return self._grid_data[self._row_index][col]

    def __iter__(self):

        return self._grid_data[self._row_index].__iter__()

    def __contains__(self, item):

        return self._grid_data[self._row_index].__contains__(item)

    bottom = property(lambda self: self.top + self.height)
    components = property(lambda self: tuple(comp for comp in self if comp is not None))
    icomponents = property(lambda self: (comp for comp in self if comp is not None))
    is_adaptable = property(lambda self: self._height is None)
    is_first = property(lambda self: self._row_index is 0)
    is_last = property(lambda self: self._row_index is len(self._grid.rows)-1)
    top = property(lambda self: self._top)

    def _update_height(self):
        """
        Adapt the row height to the required height
        """
        h = self.height
        if self.components and self.components[0].window and self.components[0].window[3] == h:
            if debug_with_assert:
                for comp in self.icomponents: assert self.components[0].window[3] == h

        top = self.top
        for comp in self.icomponents:
            col = self._grid.get_col(comp.col)
            comp.set_window((col.left, top, col.width, h), follow_movements=True)
            if comp.sticky is not None:
                comp.origin.unlock()
                comp.origin.config(pos=getattr(self.get_cell(comp.col).get_rect(), comp.sticky))
                comp.origin.lock()

        if not self.is_last:
            self.get_next_row()._update_top(top + h)

    def _update_top(self, top):

        dy = top - self.top
        if dy is 0: return
        self._top = top
        for comp in self.icomponents:
            comp.origin.unlock()
            comp.move(dy=dy)
            comp.origin.lock()

        if not self.is_last:
            self.get_next_row()._update_top(top + self.height)

    def is_empty(self):

        for cell in self:
            if cell is not None:
                return False
        return True

    def get_cell(self, col):
        """
        Return the cell whose column is indexed by 'col'
        """
        return Cell(self._grid, self._row_index, col)

    def get_cell_rect(self, col):
        """
        Optimized shortcut instead of row.get_cell(...).get_rect()
        """
        return self._grid.get_col(col).left, self.top, self._grid.get_col(col).width, self.height

    def get_height(self):

        if self._height is not None:
            return self._height

        if self.is_empty():
            return 0

        return max(comp.rect.h for comp in self if comp is not None)

    height = property(get_height)

    def get_next_row(self):

        if self.is_last: return
        return self._grid.get_row(self._row_index+1)

    def get_previous_row(self):
        """
        Return the row above self
        """
        if self.is_first: return None
        return self._grid.get_row(self._row_index-1)

    def get_rect(self):
        """
        Return the row's position (relative to the container) and size inside a rectstyle object
        """
        return self._grid_layer.container.left, self.top, self._grid_layer.container.w, self.height

    def set_height(self, height):
        """
        Set the row's height
        If height is None, the row will adapt to its hightest component
        """
        assert (height is None) or isinstance(height, (int, float)) and height >= 0
        if not self._grid.rows_are_adaptable:
            raise PermissionError("This grid has a fixed row_height : {}".format(self._grid.row_height))
        self._height = height
        self._update_height()


class GridLayer(Layer):
    """
    A GridLayer is a Layer who places its components itself, depending on their
    attributes 'row' and 'col'

    If nbrows is None, adding a component will create missing rows if needed
    nbcols works the same

    A GridLayer dimension (row or column) have two implicit modes : adaptable and fixed
    If the dimension size (col_width or row_height) is set, the mode is fixed. Else,
    it is adaptable.
    Fixed means all cell have the same dimension size
    Adaptable means cells receive their dimension size from the largest cell in the
    dimension.
    Note that for one dimension, you can reset its size. It means yous can, in the
    same grid, have an adaptable row above a fixed row, as below :

    gl = GridLayer(some_parent, cols=None, rows=5, col_width=None, row_height=25)
    grid = gl.grid
                                       # All rows have the fixed mode
    grid.get_row(3).set_height(None)   # This row gets the adaptable mode
    grid.get_row(4).set_height(45)     # This row gets the fixed mode

    Two components can't fit in the same cell

    WARNING : manipulating a grid with multi-threading might cause issues
    """

    def __init__(self, *args, nbcols=None, nbrows=None, col_width=None, row_height=None, **kwargs):

        Layer.__init__(self, *args, **kwargs)

        if nbcols is not None: assert isinstance(nbcols, int) and nbcols > 0
        if nbrows is not None: assert isinstance(nbrows, int) and nbrows > 0

        self._layer = self
        self._nbcols = None
        self._nbrows = None
        self._col_width = col_width
        self._row_height = row_height

        self._data = [[None]]  # shared ressource initialized with one row and one column
        self._cols = TypedList(Column)
        self._rows = TypedList(Row)

        Row(self, 0)
        Column(self, 0)
        if nbcols: self.set_nbcols(nbcols)
        if nbrows: self.set_nbrows(nbrows)

        if debug_with_assert:
            try:
                assert self.nbcols == len(self._data[0]) == len(self.cols)
                assert self.nbrows == len(self._data) == len(self.rows)
            except AssertionError as e:
                raise e

    def __str__(self):
        return "{}(nbrows={}, nbcols={})".format(self.__class__.__name__, self.nbrows, self.nbcols)

    col_width = property(lambda self: self._col_width)
    cols = property(lambda self: self._cols)
    cols_are_adaptable = property(lambda self: self._col_width is None)
    nbcols = property(lambda self: self._nbcols if self._nbcols is not None else len(self._data[0]))
    nbrows = property(lambda self: self._nbrows if self._nbrows is not None else len(self._data))
    row_height = property(lambda self: self._row_height)
    rows = property(lambda self: self._rows)
    rows_are_adaptable = property(lambda self: self._row_height is None)

    def _find_place_for(self, comp):
        """
        Renvoie le topleft de la cellule qui pourrait accueillir comp
        """
        1/0
        return (0, 0)  # les coordonnees seront redefines dans GridLayer.add

        if debug_with_assert: assert comp.col is not None and comp.row is not None

        if self.nbcols-1 < comp.col:
            if self._nbcols is not None:
                raise PermissionError("Cannot add {} inside {}, its col value is too big : {}"
                                      "".format(comp, self, comp.col))
            x = self.cols[-1].right
            col = self.get_col(-1)
        else:
            x = self.cols[comp.col].left if self.col_width is None else comp.col * self.col_width
            col = self.get_col(comp.col)

        if self.nbrows-1 < comp.row:
            if self._nbrows is not None:
                raise PermissionError("Cannot add {} inside {}, its row value is too big : {}"
                                      "".format(comp, self, comp.row))
            y = self.rows[-1].bottom
            row = self.nbrows-1
        else:
            y = self.rows[comp.row].top if self.row_height is None else comp.row * self.row_height
            row = comp.col

        if comp._sticky is not None:
            cell = col.get_cell(row)
            # rect = cell.get_rect()
            # pos = getattr(cell.get_rect(), comp._sticky)
            return getattr(cell.get_rect(), comp._sticky)

        return (x, y)

    def _update_size(self, *args):

        for row in self.rows:
            row._update_height()
        for col in self.cols:
            col._update_width()

    def accept(self, comp):
        """You must define at least the row or the column in order to insert a widget in a grid layer"""
        if (comp.col is None) or (comp.row is None):
            return False
        return super().accept(comp)

    def add(self, comp):

        # TODO : solve : sticky positions are not updated when the grid is resized

        if (comp.col is None) or (comp.row is None):
            raise PermissionError(
                "You must define at least the row or the column in order to insert a widget in a GridLayer")
        try:
            super().add(comp)
            if self._nbcols is None and len(self.cols)-1 < comp.col:
                self.set_nbcols(comp.col+1)
                self._nbcols = None
            if self._nbrows is None and len(self.rows)-1 < comp.row:
                self.set_nbrows(comp.row+1)
                self._nbrows = None

            if self._data[comp.row][comp.col] is not None:
                raise PermissionError("Cannot insert {} at positon : row={}, col={}, because {} is already there"
                                      "".format(comp, comp.row, comp.col, self._data[comp.row][comp.col]))

            if debug_with_assert: assert not comp.has_locked.origin, "This should be checked in Widget.__init__()"
            # comp.move_at(key="topleft", value=self._find_place_for(comp))
            if comp.sticky is not None:

                # Here, the grid layer has to be the only one who gives a position to the widget

                pos = getattr(self.get_cell(comp.row, comp.col).get_rect(), comp.sticky)
                comp.origin.config(
                    pos=pos, location=comp.sticky, reference_comp=self.container,
                    reference_location="topleft", locked=True
                )
            else:
                pos = self.get_cell(comp.row, comp.col).get_rect().topleft
                comp.origin.config(
                    pos=pos, location="topleft", reference_comp=self.container,
                    reference_location="topleft", locked=True
                )
            # if debug_with_assert:
            #     if comp._sticky is None:
            #         assert comp.topleft == self._find_place_for(comp)

            row = self.rows[comp.row]
            col = self.cols[comp.col]
            new_h = row.is_adaptable and comp.height > row.height
            new_w = col.is_adaptable and comp.width > col.width
            self._data[comp.row][comp.col] = comp
            if new_h: row._update_height()
            if new_w: col._update_width()
            if comp.window is None:
                comp.set_window(row.get_cell_rect(comp.col))
            if debug_with_assert: assert comp.window is not None, comp

            if debug_with_assert: assert self.nbcols == len(self._data[0]) == len(self.cols)
            if debug_with_assert: assert self.nbrows == len(self._data) == len(self.rows)

            # don't need owner because, if the grid is killed,
            # it means the container is killed, so the comp is also killed
            self.connect("_update_size", comp.signal.RESIZE)
            self.connect("_update_size", comp.signal.KILL)

        except Exception as e:
            raise e

    def move(self, widget, col, row):

        assert widget in self
        if self._data[row][col] is not None:
            if self._data[row][col] is widget:
                return
            raise PermissionError("This cell is already occupied by : "+str(self._data[row][col]))

        self._data[widget.row][widget.col] = None
        self._data[row][col] = widget
        widget._col = col
        widget._row = row

        widget.origin.unlock()
        if widget.sticky is not None:
            pos = getattr(self.get_cell(widget.row, widget.col).get_rect(), widget.sticky)
        else:
            pos = self.get_cell(widget.row, widget.col).get_rect().topleft
        widget.origin.config(pos=pos, locked=True)
        widget.set_window(self.rows[widget.row].get_cell_rect(widget.col))

    def remove(self, comp):

        super().remove(comp)
        assert self._data[comp.row][comp.col] is comp
        self._data[comp.row][comp.col] = None

    def get_cell(self, row, col):

        return self.get_col(col).get_cell(row)

    def get_col(self, col_index):
        """
        Return the column from index
        """
        return self._cols[col_index]

    def get_row(self, row_index):
        """
        Return the row from index
        """
        return self._rows[row_index]

    def pack(self):
        """
        Remove all empty rows and columns
        """
        for row in tuple(self.rows):
            if row.is_empty():
                self._data.pop(row.row_index)
                self.rows.remove(row)
        for col in tuple(self.cols):
            if col.is_empty():
                for data_row in self._data:
                    data_row.pop(col.col_index)
                self.cols.remove(col)

    def set_col_width(self, width):

        assert (width is None) or isinstance(width, (int, float)) and width >= 0

        self._col_width = None
        for col in self.cols:
            col.set_width(width)
        self._col_width = width

    def set_nbcols(self, nbcols):
        """
        Set the number of columns
        If nbcols is None, adding a component will create missing columns if needed
        """
        assert isinstance(nbcols, int) and nbcols > 0, nbcols
        if nbcols is not None:
            nbnew = nbcols - self.nbcols
            if nbnew < 0:
            # Delete extra columns
                if debug_with_assert: assert nbnew > 0, "A grid must have at least 1 column"
                for i in range(nbcols, self.nbcols):
                    if not self._cols[i].is_empty():
                        raise PermissionError("Cannot reduce nbcols, at least one extra column contains something")
                    self._cols[i].kill()  # TODO : implement
            elif nbnew > 0:
            # Create new columns
                for row in self._data:
                    row += [None] * nbnew
                for i in range(nbnew):
                    Column(self, i + nbcols - nbnew)
        self._nbcols = nbcols

    def set_nbrows(self, nbrows):
        """
        Set the number of rows
        If nbrows is None, adding a component will create missing rows if needed
        """
        assert isinstance(nbrows, int) and nbrows > 0
        if nbrows is not None:
            nbnew = nbrows - self.nbrows
            if nbrows < 0:
            # Delete extra columns
                if debug_with_assert: assert nbrows > 0, "A grid must have at least 1 row"
                for i in range(nbrows, self.nbrows):
                    if not self._rows[i].is_empty():
                        raise PermissionError("Cannot reduce nbrows, at least one row to delete contains something")
                    self._rows[i].kill()  # TODO : implement
            elif nbnew > 0:
            # Create new rows
                for i in range(self.nbrows, nbrows):
                    self._data += [[None] * self.nbcols]
                    Row(self, i)
        self._nbrows = nbrows

    def set_row_height(self, height):

        assert (height is None) or isinstance(height, (int, float)) and height >= 0

        self._row_height = None
        for row in self.rows:
            row.set_height(height)
        self._row_height = height

    def swap(self, widget1, widget2):

        assert widget1 in self
        assert widget2 in self

        widget1.origin.unlock()
        widget2.origin.unlock()

        self._data[widget1.row][widget1.col] = widget2
        self._data[widget2.row][widget2.col] = widget1

        widget1._col, widget1._row, widget2._col, widget2._row = widget2._col, widget2._row, widget1._col, widget1._row

        for widget in widget1, widget2:
            if widget.sticky is not None:
                pos = getattr(self.get_cell(widget.row, widget.col).get_rect(), widget.sticky)
            else:
                pos = self.get_cell(widget.row, widget.col).get_rect().topleft
            widget.origin.config(pos=pos, locked=True)
            widget.set_window(self.rows[widget.row].get_cell_rect(widget.col))







import inspect
from baopig.pybao.objectutilities import PackedFunctions, Object, TypedSet
from baopig.io.logging import LOGGER


class ApplicationExit(Exception):
    pass


class Signal:

    def __init__(self, emitter, id, catching_errors):

        assert isinstance(emitter, Communicative), emitter
        assert isinstance(id, str) and (id == id.upper()), id

        self._emitter = emitter
        self._id = id
        self._connections = []
        self._catching_errors = catching_errors
        self.emit = self.emit_with_catch if catching_errors else self.emit_no_catch

    def __str__(self):

        return f"Signal(id={self._id}, emitter={self._emitter})"

    def connect(self, command, owner=None, **options):  # TODO : rename as connect
        """
        Connect the command to the signal
        When self will emits 'signal', the owner's method 'command' will be executed
        if option 'need_arguments' is False, the emitted argumaents will be ignored
        :param command: a method of owner
        :param owner: an Communicative object
        NOTE : The "owner" parameter is very important when it comes to deletion
               When the owner is deleted, this connection is automatically killed
        """
        if not callable(command):
            raise TypeError("'{}' object is not callable".format(command))
        for con in self._connections:
            if con.slot is command:
                return
                raise PermissionError(f"{command} already connected to {self}")
        Connection(owner, self, command, **options)

    def disconnect(self, listener):
        """
        Remove all connections from this signal to the listener
        """
        for con in tuple(self._connections):
            if con.owner is listener:
                con.kill()

    def emit_with_catch(self, *args):
        """
        Emitting a signal will execute all its connected commands
        If an error occurs while executing a command, it will be logged
        The error ApplicationExit is not catched
        """

        for con in self._connections:
            try:
                con.transmit(*args) if args else con.slot()  # little optimization ?
            except ApplicationExit as e:
                raise e
            except Exception as e:
                LOGGER.warning("Error : {} -- while exectuting {}".format(e, con))

    def emit_no_catch(self, *args, catching_errors=None):
        """
        Emitting a signal will execute all its connected commands
        If an error occurs while executing a command, it will be raised
        """
        for con in self._connections:
            con.transmit(*args) if args else con.slot()  # little optimization TODO : signal.transmit_arguments

    def rem_command(self, command):

        for con in tuple(self._connections):
            if con.slot is command:
                con.kill()

    def set_catching_errors(self, catching_errors):

        self._catching_errors = bool(catching_errors)
        self.emit = self.emit_with_catch if catching_errors else self.emit_no_catch


class Signals(Object): pass


class Connection:

    def __init__(self, owner, signal, slot, **options):

        need_arguments = False
        if len(inspect.signature(slot).parameters) > 0:
            need_arguments = True
        if "need_arguments" in options:
            need_arguments = options.pop("need_arguments")

        self.owner = owner
        self.signal = signal
        self.slot = slot
        self.transmit = slot if need_arguments else lambda *args: slot()

        signal._connections.append(self)
        if owner is not None:
            owner._connections.add(self)

    def kill(self):

        self.signal._connections.remove(self)
        self.owner._connections.remove(self)


class Communicative:

    def __init__(self):

        self.signal = Signals()
        self._connections = TypedSet(Connection)

    def connect(self, method_name, signal, **options):
        """
        Connect the method called 'method_name' to the signal
        When the 'signal' will be emited, the self's method 'method_name' will be executed
        :param signal: a Signal
        :param method_name: a string representing a method of self
        """
        method = getattr(self, method_name)
        # if not inspect.ismethod(method):  # can be a decorator
        #     raise ValueError("The method_name must be a string representing a method of '{}' object"
        #                      "".format(self.__class__.__name__))
        assert isinstance(signal, Signal), signal

        signal.connect(method, owner=self, **options)

    def create_signal(self, signal_id, catching_errors=False):  # TODO : change catching_errors for testing
        # TODO : rename signal_id as name

        if not isinstance(signal_id, str) or not (signal_id == signal_id.upper()):
            raise PermissionError("A signal id must be an uppercase str")
        if hasattr(self.signal, signal_id):
            raise PermissionError("A signal already has this id : {}".format(signal_id))

        setattr(self.signal, signal_id, Signal(emitter=self, id=signal_id, catching_errors=catching_errors))

    def disconnect(self, *, signal=None, emitter=None, command=None):
        """
        Remove connections from any signal to commands owned by self
        :param signal: if set, only remove connections from this signal to commands owned by self
        :param emitter: if set, only remove connections from this emitter to commands owned by self
        """
        if signal is not None:
            signal.disconnect(self)
        elif emitter is not None:
            for signal in emitter.signal.__dict__.values():
                signal.disconnect(self)
        elif command is not None:
            for signal in self.signal:
                signal.rem_command(command)
        else:
            for con in tuple(self._connections):
                con.kill()




from .utilities import Closable
from baopig.time.timer import RepeatingTimer


class RepetivelyAnimated(Closable):
    """
    A RepetivelyAnimated object is a component who blinks every interval seconds

    WARNING : This class must be used with multiple heritance for components

    Exemple :

        class Lighthouse(Widget, RepetivelyAnimated):

    """

    def __init__(self, interval):
        """
        The component will appears and disappears every interval seconds

        :param interval: the time between appear and disappear
        """

        assert isinstance(interval, (int, float)), "interval must be a float or an integer"

        self.interval = interval

        def blink():
            if self.is_visible:
                self.hide()
            else:
                self.show()

        self.countdown_before_toggle_visibility = RepeatingTimer(interval, blink)

    def asleep(self):

        super().asleep()
        if self.countdown_before_toggle_visibility.is_paused:
            self.countdown_before_toggle_visibility.pause()

    def start_animation(self):

        if self.is_sleeping:
            self._memory.need_start_animation = True
            return

        if self.countdown_before_toggle_visibility.is_running:
            self.countdown_before_toggle_visibility.cancel()
        self.countdown_before_toggle_visibility.start()

    def stop_animation(self):

        if self.is_sleeping:
            self._memory.need_start_animation = False
            return

        self.countdown_before_toggle_visibility.cancel()

    def wake(self):

        super().wake()
        if self.countdown_before_toggle_visibility.is_paused:
            self.countdown_before_toggle_visibility.resume()




"""
Here is a few fonts availible on my computer

/Library/Fonts/Andale Mono.ttf
/Library/Fonts/Apple Chancery.ttf
/Library/Fonts/AppleGothic.ttf
/Library/Fonts/AppleMyungjo.ttf
/Library/Fonts/Arial Black.ttf
/Library/Fonts/Arial Narrow Bold Italic.ttf
/Library/Fonts/Arial Narrow Bold.ttf
/Library/Fonts/Arial Narrow Italic.ttf
/Library/Fonts/Arial Narrow.ttf
/Library/Fonts/Arial Rounded Bold.ttf
/Library/Fonts/Arial Unicode.ttf
/Library/Fonts/Ayuthaya.ttf
/Library/Fonts/BigCaslon.ttf
/Library/Fonts/Bodoni 72 Smallcaps Book.ttf
/Library/Fonts/Bodoni Ornaments.ttf
/Library/Fonts/Bradley Hand Bold.ttf
/Library/Fonts/Chalkduster.ttf
/Library/Fonts/Comic Sans MS Bold.ttf
/Library/Fonts/Comic Sans MS.ttf
/Library/Fonts/Courier New Bold Italic.ttf
/Library/Fonts/Courier New Bold.ttf
/Library/Fonts/Courier New Italic.ttf
/Library/Fonts/Courier New.ttf
/Library/Fonts/DIN Alternate Bold.ttf
/Library/Fonts/DIN Condensed Bold.ttf
/Library/Fonts/Diwan Thuluth.ttf
/Library/Fonts/Farisi.ttf
/Library/Fonts/Georgia Bold Italic.ttf
/Library/Fonts/Georgia Bold.ttf
/Library/Fonts/Georgia Italic.ttf
/Library/Fonts/Georgia.ttf
/Library/Fonts/Gungseouche.ttf
/Library/Fonts/Gurmukhi.ttf
/Library/Fonts/HeadlineA.ttf
/Library/Fonts/Herculanum.ttf
/Library/Fonts/Hoefler Text Ornaments.ttf
/Library/Fonts/Impact.ttf
/Library/Fonts/InaiMathi.ttf
/Library/Fonts/Khmer Sangam MN.ttf
/Library/Fonts/Kokonor.ttf
/Library/Fonts/Krungthep.ttf
/Library/Fonts/Lao Sangam MN.ttf
/Library/Fonts/Luminari.ttf
/Library/Fonts/Microsoft Sans Serif.ttf
/Library/Fonts/Mishafi Gold.ttf
/Library/Fonts/Mishafi.ttf
/Library/Fonts/NISC18030.ttf
/Library/Fonts/Osaka.ttf
/Library/Fonts/OsakaMono.ttf
/Library/Fonts/PCmyoungjo.ttf
/Library/Fonts/Pilgiche.ttf
/Library/Fonts/PlantagenetCherokee.ttf
/Library/Fonts/Sathu.ttf
/Library/Fonts/Silom.ttf
/Library/Fonts/Skia.ttf
/Library/Fonts/Tahoma Bold.ttf
/Library/Fonts/Tahoma.ttf
/Library/Fonts/Trattatello.ttf
/Library/Fonts/Trebuchet MS Bold Italic.ttf
/Library/Fonts/Trebuchet MS Bold.ttf
/Library/Fonts/Trebuchet MS Italic.ttf
/Library/Fonts/Trebuchet MS.ttf
/Library/Fonts/Webdings.ttf
/Library/Fonts/Zapfino.ttf
/Library/Fonts/儷宋 Pro.ttf
/Library/Fonts/儷黑 Pro.ttf
/Library/Fonts/华文仿宋.ttf
/Library/Fonts/华文细黑.ttf
/Library/Fonts/华文黑体.ttf
/Library/Fonts Disabled/Arial Bold Italic.ttf
/Library/Fonts Disabled/Arial Bold.ttf
/Library/Fonts Disabled/Arial Italic.ttf
/Library/Fonts Disabled/Arial.ttf
/Library/Fonts Disabled/Brush Script.ttf
/Library/Fonts Disabled/Times New Roman Bold Italic.ttf
/Library/Fonts Disabled/Times New Roman Bold.ttf
/Library/Fonts Disabled/Times New Roman Italic.ttf
/Library/Fonts Disabled/Times New Roman.ttf
/Library/Fonts Disabled/Verdana Bold Italic.ttf
/Library/Fonts Disabled/Verdana Bold.ttf
/Library/Fonts Disabled/Verdana Italic.ttf
/Library/Fonts Disabled/Verdana.ttf
/Library/Fonts Disabled/Wingdings 2.ttf
/Library/Fonts Disabled/Wingdings 3.ttf
/Library/Fonts Disabled/Wingdings.ttf
/Library/Fonts/Microsoft/Arial Bold Italic.ttf
/Library/Fonts/Microsoft/Arial Bold.ttf
/Library/Fonts/Microsoft/Arial Italic.ttf
/Library/Fonts/Microsoft/Arial.ttf
/Library/Fonts/Microsoft/Batang.ttf
/Library/Fonts/Microsoft/Bookshelf Symbol 7.ttf
/Library/Fonts/Microsoft/Brush Script.ttf
/Library/Fonts/Microsoft/Calibri Bold Italic.ttf
/Library/Fonts/Microsoft/Calibri Bold.ttf
/Library/Fonts/Microsoft/Calibri Italic.ttf
/Library/Fonts/Microsoft/Calibri.ttf
/Library/Fonts/Microsoft/Cambria Bold Italic.ttf
/Library/Fonts/Microsoft/Cambria Bold.ttf
/Library/Fonts/Microsoft/Cambria Italic.ttf
/Library/Fonts/Microsoft/Cambria Math.ttf
/Library/Fonts/Microsoft/Cambria.ttf
/Library/Fonts/Microsoft/Candara Bold Italic.ttf
/Library/Fonts/Microsoft/Candara Bold.ttf
/Library/Fonts/Microsoft/Candara Italic.ttf
/Library/Fonts/Microsoft/Candara.ttf
/Library/Fonts/Microsoft/Consolas Bold Italic.ttf
/Library/Fonts/Microsoft/Consolas Bold.ttf
/Library/Fonts/Microsoft/Consolas Italic.ttf
/Library/Fonts/Microsoft/Consolas.ttf
/Library/Fonts/Microsoft/Constantia Bold Italic.ttf
/Library/Fonts/Microsoft/Constantia Bold.ttf
/Library/Fonts/Microsoft/Constantia Italic.ttf
/Library/Fonts/Microsoft/Constantia.ttf
/Library/Fonts/Microsoft/Corbel Bold Italic.ttf
/Library/Fonts/Microsoft/Corbel Bold.ttf
/Library/Fonts/Microsoft/Corbel Italic.ttf
/Library/Fonts/Microsoft/Corbel.ttf
/Library/Fonts/Microsoft/Franklin Gothic Book Italic.ttf
/Library/Fonts/Microsoft/Franklin Gothic Book.ttf
/Library/Fonts/Microsoft/Franklin Gothic Medium Italic.ttf
/Library/Fonts/Microsoft/Franklin Gothic Medium.ttf
/Library/Fonts/Microsoft/Gabriola.ttf
/Library/Fonts/Microsoft/Gill Sans MT Bold Italic.ttf
/Library/Fonts/Microsoft/Gill Sans MT Bold.ttf
/Library/Fonts/Microsoft/Gill Sans MT Italic.ttf
/Library/Fonts/Microsoft/Gill Sans MT.ttf
/Library/Fonts/Microsoft/Gulim.ttf
/Library/Fonts/Microsoft/Lucida Console.ttf
/Library/Fonts/Microsoft/Lucida Sans Unicode.ttf
/Library/Fonts/Microsoft/Marlett.ttf
/Library/Fonts/Microsoft/Meiryo Bold Italic.ttf
/Library/Fonts/Microsoft/Meiryo Bold.ttf
/Library/Fonts/Microsoft/Meiryo Italic.ttf
/Library/Fonts/Microsoft/Meiryo.ttf
/Library/Fonts/Microsoft/MS Gothic.ttf
/Library/Fonts/Microsoft/MS Mincho.ttf
/Library/Fonts/Microsoft/MS PGothic.ttf
/Library/Fonts/Microsoft/MS PMincho.ttf
/Library/Fonts/Microsoft/MS Reference Sans Serif.ttf
/Library/Fonts/Microsoft/MS Reference Specialty.ttf
/Library/Fonts/Microsoft/Palatino Linotype Bold Italic.ttf
/Library/Fonts/Microsoft/Palatino Linotype Bold.ttf
/Library/Fonts/Microsoft/Palatino Linotype Italic.ttf
/Library/Fonts/Microsoft/Palatino Linotype.ttf
/Library/Fonts/Microsoft/Perpetua Bold Italic.ttf
/Library/Fonts/Microsoft/Perpetua Bold.ttf
/Library/Fonts/Microsoft/Perpetua Italic.ttf
/Library/Fonts/Microsoft/Perpetua.ttf
/Library/Fonts/Microsoft/PMingLiU.ttf
/Library/Fonts/Microsoft/SimSun.ttf
/Library/Fonts/Microsoft/Times New Roman Bold Italic.ttf
/Library/Fonts/Microsoft/Times New Roman Bold.ttf
/Library/Fonts/Microsoft/Times New Roman Italic.ttf
/Library/Fonts/Microsoft/Times New Roman.ttf
/Library/Fonts/Microsoft/Tw Cen MT Bold Italic.ttf
/Library/Fonts/Microsoft/Tw Cen MT Bold.ttf
/Library/Fonts/Microsoft/Tw Cen MT Italic.ttf
/Library/Fonts/Microsoft/Tw Cen MT.ttf
/Library/Fonts/Microsoft/Verdana Bold Italic.ttf
/Library/Fonts/Microsoft/Verdana Bold.ttf
/Library/Fonts/Microsoft/Verdana Italic.ttf
/Library/Fonts/Microsoft/Verdana.ttf
/Library/Fonts/Microsoft/Wingdings 2.ttf
/Library/Fonts/Microsoft/Wingdings 3.ttf
/Library/Fonts/Microsoft/Wingdings.ttf
/System/Library/Fonts/Apple Braille Outline 6 Dot.ttf
/System/Library/Fonts/Apple Braille Outline 8 Dot.ttf
/System/Library/Fonts/Apple Braille Pinpoint 6 Dot.ttf
/System/Library/Fonts/Apple Braille Pinpoint 8 Dot.ttf
/System/Library/Fonts/Apple Braille.ttf
/System/Library/Fonts/Apple Color Emoji.ttf
/System/Library/Fonts/Keyboard.ttf
/System/Library/Fonts/LastResort.ttf
/System/Library/Fonts/Symbol.ttf
/System/Library/Fonts/ZapfDingbats.ttf
"""


import glob
import pygame
from baopig.pybao.issomething import is_color
from pygame.locals import *
from baopig.ressources import *
from baopig.io import LOGGER
from baopig.ressources import ressources


class Font:

    def __init__(self,
        file=None,
        height=None,
        color=None,
        bold=None,
        italic=None,
        underline=None,
        text_owner=None
    ):

        if file is None: file = ressources.font.file
        if height is None: height = ressources.font.height
        if color is None: color = ressources.font.color

        assert file is None or isinstance(file, str)

        self._file = None
        self._filepath = None
        self._height = None
        self._ascent = None
        self._font = None
        self._color = None
        self._text_owner_wkref = (lambda: None) if text_owner is None else text_owner.get_weakref()

        self.config(file=file, height=height, color=color, bold=bold, italic=italic, underline=underline, update_owner=False)

        # for i in range(30):
        #     f = pygame.font.Font(self.file, i)
        #     print(i, f.size(" ")[1])

    def __str__(self):

        return f"Font(height:{self.height}, ascent:{self._ascent}, file={self.file})"
    __repr__ = __str__

    color = property(lambda self: self._color)
    file = property(lambda self: self._file)
    filepath = property(lambda self: self._filepath)
    height = property(lambda self: self._height)
    text_owner = property(lambda self: self._text_owner_wkref())

    def config(self, file=None, height=None, color=None, bold=None, italic=None, underline=None, update_owner=True):

        if file is not None:
            self._file = file

            if file == "monospace":
                file = "JetBrainsMono-Medium.ttf"
            elif not file.endswith(".ttf"):
                if file.count("."):
                    LOGGER.warning("Can only look for tff fonts, not " + file)
                    self._file = file = ressources.font.file
                else:
                    file += ".ttf"

            if file not in filepaths:
                # print("get_filepath(", file, end=") -> ", sep="")
                filepaths[file] = get_filepath(file)
                # print(filepaths[file])
            self._filepath = filepaths[file]

        if height is not None:
            assert height > 1, "A font height must be higher than 1"
            self._height = height
        if file is not None or height is not None:

            """if self.height in _all[file]._fonts:
                ascent = _all[file].get_font(self.height)._ascent
            else:
                ascent = self.height
                all_from_file = _all[self.file]

                little_brother = self.height - 1
                if not all_from_file._fonts:
                    little_brother = 0
                while little_brother > 0 and little_brother not in all_from_file._fonts:
                    little_brother -= 1
                if little_brother > 0:
                    ascent = all_from_file.get_font(little_brother)._ascent

                elif all_from_file._fonts and little_brother == 0:
                    big_brother = self.height + 1
                    while big_brother not in all_from_file._fonts:
                        big_brother += 1
                    ascent = all_from_file.get_font(big_brother)._ascent
                    while height < all_from_file.get_font(ascent=ascent).size("")[1]:
                    while height < pygame.font.Font(self.font, ascent).size("")[1]:
                        ascent -= 1
                        if ascent <= 0:
                            raise PermissionError("Too small value for font height : " + str(self.height))

                while self.height < pygame.font.Font(self.file, ascent).size("")[1]:
                    ascent -= 1
                while self.height > pygame.font.Font(self.file, ascent).size("")[1]:
                    ascent += 1
                if ascent <= 0:
                    raise PermissionError("Too small value for font height : " + str(self.height))
                
            self._ascent = ascent
            self._font = pygame.font.Font(self.file, self._ascent)"""

            self._font = _all[self.filepath].get_font(height=self.height)
            self._ascent = self._font.get_ascent()

        if color is not None:
            try:               color = pygame.Color(color)
            except ValueError: color = pygame.Color(*color)
            self._color = color
        if bold is not None:
            self._font.set_bold(bold)
        if italic is not None:
            self._font.set_italic(italic)
        if underline is not None:
            self._font.set_underline(underline)

        if update_owner and self.text_owner is not None:
            self.text_owner.set_text(self.text_owner.text)

    def copy(self, text_owner=None):

        return Font(
            file=self.file,
            height=self.height,
            color=self.color,
            bold=self._font.get_bold(),
            italic=self._font.get_italic(),
            underline=self._font.get_underline(),
            text_owner=text_owner
        )

    def get_width(self, text):

        return self._font.size(text)[0]

    def render(self, text, background_color=None):

        assert background_color is None or len(background_color) == 3, \
            "Don't like rendering a text with a transparent background color"

        rendering = self._font.render(text, 1, self._color, background_color)
        if rendering.get_height() != self._height:
            assert rendering.get_height() > self._height
            rendering = rendering.subsurface((0, 0, rendering.get_width(), self._height)).copy()
        return rendering


# NOTE : use TypedDict(pygame.font.Font)
class _Fonts:
    """
    Pour plus de clarte, il a ete decide que Font(30) cree une ploice de taille 30
    En effet, pygame.font.Font(30) cree une police de hauteur 21
    Afin de trouver efficacement quelle valeur faut-il donner à pygame.font.Font
    pour obtenir une police de taille 30, on enregistre les polices auparavent crees
    dans _fonts

    Exemple :

    """

    def __init__(self, filepath):

        if filepath == "default":
            filepath = None
        pygame.font.init()
        self._fonts = {}
        self._fonts_by_ascent = {}
        self._filepath = filepath

    def get_font(self, height=None, ascent=None):
        """
        Methode qui renvoie une police de hauteur 'height' en pixels ou une police cree
        avec la hauteur bizarre de pygame, appelee ici 'ascent'
        """
        if ascent is None:
            assert isinstance(height, int), "height must be an integer (it's a font size)"
            assert height > 0, "A text must have a positive height"

            try:
                # Si le dictionnaire des polices contient deja la taille de police demandee
                return self._fonts[height]
            except KeyError:
                pass

            # Sinon on l'ajoute au dictionnaire
            ascent = height

            while height < self.get_font(ascent=ascent).get_height():
                ascent -= 1
            while height > self.get_font(ascent=ascent).get_height():
                ascent += 1
            if ascent <= 0:
                raise PermissionError("Too small value for font height : " + str(height))

            try:
                # Si le dictionnaire des polices contient deja la taille de police demandee
                return self._fonts[height]
            except KeyError:
                pass

            if self.get_font(ascent=ascent).get_height() < height:
                # print(self._fonts, self.get_font(ascent=ascent).get_height(), height)
                raise AssertionError
            slightly_higher_font = self.get_font(ascent=ascent)
            for i in range(self.get_font(ascent=ascent).get_height() - height):
                self._fonts[height+i+1] = slightly_higher_font
            self._fonts_by_ascent[ascent] = slightly_higher_font
            return slightly_higher_font

        else:
            assert height is None

            try:
                # Si le dictionnaire des polices contient deja la taille de police demandee
                return self._fonts_by_ascent[ascent]
            except KeyError:
                pass

            new_font = pygame.font.Font(self._filepath, ascent)
            self._fonts[new_font.get_height()] = new_font
            self._fonts_by_ascent[ascent] = new_font
            return new_font


class _All(dict):
    def __getitem__(self, filepath):
        if filepath not in self:
            self[filepath] = _Fonts(filepath)
        return super().__getitem__(filepath)
_all = _All()


def get_filepath(file):
    import os
    exec_dir = os.getcwd()
    for path in glob.iglob(exec_dir + "/*.ttf"):
        if path.endswith(file): return path
    for path in glob.iglob(exec_dir + "/*/*.ttf"):
        if path.endswith(file): return path
    for path in glob.iglob(exec_dir + "/*/*.ttf"):
        if path.endswith(file): return path
    for path in glob.iglob(exec_dir + "/*/*/*.ttf"):
        if path.endswith(file): return path
    for path in glob.iglob(__file__[:-7] + "lib/*.ttf"):
        if path.endswith(file): return path
    for path in glob.iglob(__file__[:-7] + "lib/*/*.ttf"):
        if path.endswith(file): return path
    for path in glob.iglob(__file__[:-7] + "lib/*/*/*.ttf"):
        if path.endswith(file): return path
    for path in glob.iglob("/System/Library/Fonts/*.ttf"):
        if path.endswith(file): return path
    for path in glob.iglob("/System/Library/Fonts/*.ttf"):
        if path.endswith(file): return path
    for path in glob.iglob("/Library/Fonts/Microsoft/*.ttf"):
        if path.endswith(file): return path
    for path in glob.iglob("/Library/Fonts Disabled/*.ttf"):
        if path.endswith(file): return path
    for path in glob.iglob("/Library/Fonts/*.ttf"):
        if path.endswith(file): return path
    LOGGER.warning("Font not found : {} (the font {} will be used instaed)".format(file, ressources.font.file))


filepaths = {}


if __name__ == "__main__":
    import os
    exec_dir = os.getcwd()
    for path in glob.iglob(__file__[:-7] + "lib/*.ttf"):
        print(path)
    for path in glob.iglob(__file__[:-7] + "lib/*/*.ttf"):
        print(path)
    for path in glob.iglob(__file__[:-7] + "lib/*/*/*.ttf"):
        print(path)
    for path in glob.iglob("/System/Library/Fonts/*.ttf"):
        print(path)
    for path in glob.iglob("/System/Library/Fonts/*.ttf"):
        print(path)
    for path in glob.iglob("/Library/Fonts/Microsoft/*.ttf"):
        print(path)
    for path in glob.iglob("/Library/Fonts Disabled/*.ttf"):
        print(path)
    for path in glob.iglob("/Library/Fonts/*.ttf"):
        print(path)

    h = 100
    font = Font('monospace', height=h)
    assert font.height == h == font._font.get_height()
    print(font)




from .style         import Theme, DarkTheme  # TODO : import DarkTheme and all pre-defined themes
from .runable       import Runable
from .hoverable     import Hoverable
from .utilities     import *
from .animations    import *
from .widget        import Widget, Communicative
from .resizable     import ResizableWidget
from .animatedwidget import AnimatedWidget
from .image         import *
from .shapes        import *
from .layer         import Layer, TemporaryLayer
from .gridlayer     import GridLayer
from .layersmanager import LayersManager
from .container     import Container
from .scrollable    import Scrollable, ScrollableByMouse
from .box           import Box
from .zone          import Zone, SubZone
from .selections    import Selectable, Selector
from .scene         import Scene
from .application   import Application




import pygame
from baopig._lib import Rectangle, Runable


class ProgressBar(Rectangle, Runable):

    def __init__(self, parent, min, max, get_progress, *args, **kwargs):

        try: Rectangle.__init__(self, parent, *args, **kwargs)
        except AttributeError: pass  # 'ProgressBar' object has no attribute '_progression'
        Runable.__init__(self)

        # Non protected fields (don't need it)
        self.min = min
        self.max = max
        self.get_progress = get_progress
        # Protected field
        self._progression = 0  # percentage between 0 and 1

        self.set_border((0, 0, 0), 2)  # TODO : ressource
        self.run()
        self.start_running()

    progression = property(lambda self: self._progression)

    def paint(self):
        """
        If size is set, this method resizes the ProgressBar
        """
        # size = size if size is not None else self.size
        # surface = pygame.Surface(size, pygame.SRCALPHA)
        self.surface.fill((0, 0, 0, 0))
        pygame.draw.rect(self.surface, self.color, (0, 0, self.progression * size[0], size[1]))
        pygame.draw.rect(self.surface, self.border_color, (self.auto_hitbox), self.border_width * 2 - 1)
        # self.set_surface(surface)

    def run(self):
        self._progression = (float(self.get_progress()) - self.min) / (self.max - self.min)
        self.send_paint_request()





import pygame
from string import printable as string_printable
from baopig.pybao.issomething import *
from baopig.pybao.objectutilities import TypedDeque, PrefilledFunction
from baopig.ressources import *
from baopig.io import LOGGER, mouse
from baopig._lib import Closable, Focusable, Container, Selector, Rectangle
from .text import Text
from .label import DynamicLabel


class InputBox(DynamicLabel, Selector):
    """
    Une InputBox est un Container.
    Elle contient un DynamicLabel dont le texte est saisi au clavier par l'utilisateur
    Les entrees peuventent etre filtrees grace au parametre char_filter : seuls les
    caracteres contenus dans char_filter sont acceptes comme entrees
    Afin d'ecouter les entrees du clavier, la InputBox a besoin d'avoir le focus
    """

    class HistoricElement:
        # TODO : make lighter ? via a single string concatening both values ? via a tuple inheritance ?
        def __init__(self, text, cursor_index):

            assert isinstance(text, str)
            assert isinstance(cursor_index, int)

            self.text = text
            self.cursor_index = cursor_index

    def __init__(self,
                 parent,
                 size=None,
                 pos=None,
                 command=None,
                 exec_command_on_defocus=False,
                 presentation_text=None,
                 font_size=None,
                 font_color=None,
                 background_color=None,
                 default_text=None,
                 entry_type=None,
                 char_filter=None,
                 maxlen=None,
                 name=None):

        if size is None:
            h = ressources.font.height
            size = (h * 5, h + int(h / 6) * 2)
        if command is None: command = lambda: None
        if presentation_text is None: presentation_text = "Enter your text here"
        if font_color is None: font_color = TEXTBOXES_FONT_COLOR
        if background_color is None: background_color = TEXTBOXES_BG_COLOR
        if default_text is None: default_text = "Unset"  # If we don't precise something, it will be "Unset"
        if entry_type is None: entry_type = str

        # TODO : Faire un background uni et ajouter deux Rectangles pour le contour (un pour la ligne noire, l'autre pour marquer le margin)
        background = pygame.Surface(size, pygame.SRCALPHA)
        background.fill(background_color)
        pygame.draw.rect(background, TEXTBOXES_OUTLINE_COLOR, (0, 0) + size, 3)  # Contour

        assert isinstance(default_text, str)
        assert entry_type in (str, float, int), \
                    "Entries of InputBox can only be of type str, float or int"
        if char_filter is not None:
            # char_filter peut etre un string ou une liste de caracteres
            assert hasattr(char_filter, "__iter__"), "char_filter must be an iterable"
            for elt in char_filter:
                assert isinstance(elt, str), "char_filter must only contains characters"
                assert len(elt) == 1, "char_filter must only contains characters"

        # Inheritance
        DynamicLabel.__init__(
            self,
            parent=parent,
            pos=pos,
            text=default_text,
            size=size,
            margin=int(size[1] / 6),
            font_size=font_size,
            font_color=font_color,
            background_color=background_color,
            # TODO : un mode qui permet d'avoir une hitbox fixe et un font_size fixe tout en permettant les changements sur text
            modification_mode=FIXED_HITBOX,
            text_location="midleft",
            name=name
        )
        Selector.__init__(self)
        # INFO : InputBox is has stopped running because of static text

        def handle_click():
            # self.selection.start(self.text_component.find_mouse_index())
            # if self.cursor.is_sleeping:
            #     self.cursor.wake()
            # self.cursor.set_index(self.selection.index_end)
            print("CLICK")
        # self.handle_click.add(handle_click)
        # self.handle_drag.add(lambda: self.cursor.set_index(self.selection.index_end))

        def handle_focus():
            if self.cursor.is_sleeping:
                index = self.text_component.find_mouse_index() if mouse.button[1] else len(self.text)
                self.cursor.set_index(index)
            # else, the cursor have been woke up by a click
            print("FOCUS")
        def handle_defocus():
            self.cursor.asleep()
        def handle_enter():
            # TODO : use handle.add(...)
            assert self.is_focused

            if not self.text:
                if self.default_text:
                    self.set_text(self.default_text, cursor_index=len(self.default_text))

            mouse.defocus(self)  # -> generate a handle_defocus -> del cursor
                                 #                              -> command() (optionnal)

            if not exec_command_on_defocus:
                try:
                    command(self.entry_type(self.text))  # TODO : InputNumber and InputText
                except ValueError:
                    LOGGER.warning("Wrong value : {}, expected object of {} type".format(self.text, self.entry_type))
                except TypeError:  # missing 1 required positional argument: 'value' ?
                    command()
                except Exception as e:
                    raise e

        self.handle_focus.add(handle_focus)
        self.handle_defocus.add(handle_defocus)
        # self.handle_defocus.add(self.selection.close)
        self.handle_enter.add(handle_enter)
        # self.handle_enter.add(handle_enter)
        if exec_command_on_defocus:
            self.handle_defocus.add(command)

        # Historic
        max_item_stored = 5
        self.historic = TypedDeque(InputBox.HistoricElement, maxlen=max_item_stored)
        self.back_historic = TypedDeque(InputBox.HistoricElement, maxlen=max_item_stored)

        # Le contenu du label
        self._presentation_text_ref = Text(
            parent=self,
            text=presentation_text,
            pos=self.text_component.topleft,
            font_height=self.text_component.font.height,
            font_color=([255 - int((255 - font_color[i]) * .75)
                         for i in range(3)]),  # Le texte de presentation est plus clair
            name=self.name + " -> presentation_text",
        ).get_weakref()
        self.presentation_text.set_nontouchable()
        if default_text:
            self.presentation_text.hide()

        self.default_text = default_text
        self.char_filter = char_filter
        self.entry_type = entry_type
        self.maxlen = maxlen
        self.pos_label = 0  # Le decalage vers la gauche du label, lorsque le texte est trop long

        # Save the initial state
        self.save()

        margin_rect = Rectangle(
            parent=self,
            pos=(0, 0),
            size=self.size,
            border_width=self.margin,
            color=self.background.color,
            name=self.name + " -> margin"
        )
        margin_rect.set_nontouchable()
        outline_rect = Rectangle(
            parent=self,
            pos=(0, 0),
            size=self.size,
            border_width=2,
            color=TEXTBOXES_OUTLINE_COLOR,
            name=self.name + " -> outline"
        )
        outline_rect.set_nontouchable()

        """
        Tentatives de recuperer le presse papier

        pygame.scrap.init()
        if pygame.scrap.lost ():
            print("No content from me anymore. The clipboard is used by someone else.")
        pygame.scrap.put("own_data", "Bonjour")
        print(pygame.scrap.get("own_data"))
        types = pygame.scrap.get_types()
        #for type in types:
            #print(type, pygame.scrap.get(type))
        """

    have_selection = property(lambda self: self.selection.is_visible)
    presentation_text = property(lambda self: self._presentation_text_ref())

    def accept(self, new_char):
        """
        Test if a given character can be added to the text
        """

        if not new_char in string_printable: return False
        try:
            self.entry_type(new_char)
        except ValueError as e:
            if self.entry_type == float and new_char == '.':
                # Prevent apparition of two points in a float
                return new_char not in self.text
            else:
                return False
        return self.char_filter is None or new_char in self.char_filter

    def del_selected_text(self):
        """
        Supprime les caracteres seletionnes
        """

        if self.have_selection:
            # We use sorted because the self.index_end can be lower than self.index_start
            index_start, index_end = self.selection.sorted
            self.set_text(self.text[:index_start] + self.text[index_end:], cursor_index=index_start)
            self.selection.close()

    def handle_KEYDOWN(self, event):

        print("handle_KEYDOWN", event.key, self.is_focused)
        if self.is_focused:
            self.write(event)

    def redo(self):
        """
        Restaure la derniere modification
        """
        if len(self.back_historic):

            backup = self.back_historic.pop()  # last element of self.back_historic, the current state
            self.historic.append(backup)

            self.set_text(backup.text, cursor_index=backup.cursor_index, save=False)
        else:
            LOGGER.debug("Cannot redo last operation because the operations historic is empty")

    def save(self):

        self.historic.append(InputBox.HistoricElement(self.text, self.cursor.index))
        self.back_historic.clear()

    def set_text(self, text, cursor_index=None, save=True):

        if self.maxlen is not None and len(text) > self.maxlen:
            return

        super().set_text(text)
        if cursor_index is not None: self.cursor.set_index(cursor_index)
        if save: self.save()

        if not text and not self.presentation_text.is_visible:
            self.presentation_text.show()
        if text and self.presentation_text.is_visible:
            self.presentation_text.hide()

    def undo(self):
        """
        Annule la derniere modification
        """
        if len(self.historic) > 1:

            backup = self.historic.pop()  # last element of self.historic, which is the state before undo()
            self.back_historic.append(backup)

            last_state = self.historic[-1]
            self.set_text(last_state.text, cursor_index=last_state.cursor_index, save=False)
        else:
            LOGGER.debug("Cannot undo last operation because the operations historic is empty")

    def write(self, event):
        """
        N'accepte que les evenements du clavier
        Si la touche est speciale, effectue sa propre fonction
        Modifie le placement du curseur

        exemple d'event accepte:
            <Event(2-KeyDown {'unicode': '3', 'key': 34, 'mod': 1, 'scancode': 20, 'window': None})>
        """

        assert isinstance(event, pygame.event.EventType)
        assert hasattr(event, "key"), "write() need a keyboard entry"

        # Maj + ...
        if event.mod in (1, 2):
            if event.key == pygame.K_LEFT:
                # On avance la selection vers la gauche
                if not self.have_selection:
                    self.selection.start(self.cursor.index)
                self.cursor.set_index(self.cursor.index - 1)
                self.selection.index_end = self.cursor.index
            elif event.key == pygame.K_RIGHT:
                # On avance la selection vers la droite
                if not self.have_selection:
                    self.selection.start(self.cursor.index)
                self.cursor.set_index(self.cursor.index + 1)
                self.selection.index_end = self.cursor.index

        # Cmd + ...
        elif event.mod in (1024, 2048):
            if event.key == pygame.K_a:  # Select all
                self.selection.start(0)
                self.selection.index_end = len(self.text)
                self.cursor.set_index(self.selection.index_end)
            elif event.key == pygame.K_z:
                self.undo()
        # TODO : Cmd + Fleche works same as Fn + Fleche
        # TODO : Alt + Fleche go to word side

        # Maj + Cmd + ...
        elif event.mod in (1025, 1026, 2049, 2050):
            if event.key == pygame.K_z:
                self.redo()

        # Fleches
        elif event.key in (pygame.K_LEFT, pygame.K_HOME, pygame.K_RIGHT, pygame.K_END):
            if self.have_selection:
                if event.key in (pygame.K_LEFT, pygame.K_HOME):
                    self.cursor.set_index(self.selection.lowest)
                elif event.key in (pygame.K_RIGHT, pygame.K_END):
                    self.cursor.set_index(self.selection.highest)
                self.selection.close()
            elif event.key == pygame.K_LEFT:
                self.cursor.set_index(self.cursor.index - 1)
            elif event.key == pygame.K_HOME:  # Fn + K_LEFT
                self.cursor.set_index(0)
            elif event.key == pygame.K_RIGHT:
                self.cursor.set_index(self.cursor.index + 1)
            elif event.key == pygame.K_END:  # Fn + K_RIGHT
                self.cursor.set_index(len(self.text))

        # Suppression
        elif event.key == pygame.K_BACKSPACE:
            if self.have_selection:
                self.del_selected_text()
            elif self.cursor.index > 0:
                self.set_text(self.text[:self.cursor.index - 1] + self.text[self.cursor.index:],
                              cursor_index=self.cursor.index - 1)

        # Validation
        elif event.key in (pygame.K_RETURN, pygame.K_TAB):
            # because RETURN and TAB are in string.printable, we need to prevent writing them
            return

        # Lettres (minuscules et majuscules)
        if event.unicode and self.accept(event.unicode):
            if self.have_selection: self.del_selected_text()
            self.set_text(self.text[:self.cursor.index] + event.unicode + self.text[self.cursor.index:],
                          cursor_index=self.cursor.index + 1)
            # print(event)





import pygame
from .utilities import MarginType, Color, paint_lock
from .layer import Layer
from .container import Container
from .image import Image


class BoxRect(pygame.Rect):

    def __init__(self, rect, margin, out=False):

        if out:
            pygame.Rect.__init__(
                self,
                rect.left - margin.left,
                rect.top - margin.top,
                rect.width + margin.left + margin.right,
                rect.height + margin.top + margin.bottom
            )
        else:
            pygame.Rect.__init__(
                self,
                rect.left + margin.left,
                rect.top + margin.top,
                rect.width - margin.left - margin.right,
                rect.height - margin.top - margin.bottom
            )


class Box(Container):

    STYLE = Container.STYLE.substyle()
    STYLE.create(
        margin = 0,
        border_width = 0,
        padding = 0,
        background_image = None
    )
    STYLE.set_type("margin", MarginType)
    STYLE.set_type("border_width", MarginType)
    STYLE.set_type("padding", MarginType)

    # NOTE : if width or height is defined in style, and a background_image is set,
    # the width and height values will be ignored

    def __init__(self, parent, **options):

        self.inherit_style(parent, options)  # TODO : **options ?

        Container.__init__(self, parent, **options)

        self._margin = self.style["margin"]
        self._border = self.style["border_width"]
        self._padding = self.style["padding"]
        self.background_layer = None
        self._background_ref = lambda: None

        self.connect("handle_resize", self.signal.RESIZE)
        background_image = self.style["background_image"]
        if background_image is not None:
            self.set_background_image(background_image)
            return
            if self.style["width"] is None:
                self.style.modify(width=background_image.get_width())
            if self.style["height"] is None:
                self.style.modify(height=background_image.get_height())

    background = property(lambda self: self._background_ref())
    border = property(lambda self: self._border)
    border_rect = property(lambda self: self.rect)
    content_rect = property(lambda self: BoxRect(self.padding_rect, self.padding))
    margin = property(lambda self: self._margin)
    margin_rect = property(lambda self: BoxRect(self.rect, self.margin, out=True))
    padding = property(lambda self: self._padding)
    padding_rect = property(lambda self: BoxRect(self.rect, self.border))

    def handle_resize(self, old_size):

        if self.background is not None:
            self.background.resize(*self.size)

    def set_background_image(self, surf, background_adapt=True):
        """
        If background_adapt is True, the surf adapts to the zone's size
        Else, the zone's size adapts to the background_image
        """
        if surf is None:
            if self.background is not None:
                with paint_lock:
                    self.background.kill()
            return
        if background_adapt and surf.get_size() != self.size:
            surf = pygame.transform.scale(surf, self.size)
        if self.background_layer is None:
            self.background_layer = Layer(self, Image, name="background_layer", level=self.layers_manager.BACKGROUND)
        with paint_lock:
            if self.background is not None:
                self.background.kill()
                assert self.background is None
            self._background_ref = Image(self, surf, pos=(0, 0), layer=self.background_layer).get_weakref()
            if background_adapt is False:
                self.resize(*self.background.size)

            def handle_background_kill(weakref):
                if weakref is self._background_ref:
                    self._background_ref = lambda: None
            self.background.signal.KILL.connect(handle_background_kill)




"""
Welcome to BaoPig

BAOPIG : Boite A Outils Pour Interfaces Graphiques

"""

# TODO : compilation executable
from .version import version
print("Hello, this is baopig version", version)
from pygame import *
from .pybao.issomething import *
from .pybao.objectutilities import Object, PrefilledFunction, PackedFunctions, \
                                         TypedDict, TypedList, TypedDeque, TypedSet

from .ressources import *
from .io import *
from .time import *
from ._lib import *
from .widgets import *

display = None  # protection for pygame.display

__version__ = str(version)


def debug_with_logging():

    LOGGER.add_debug_filehandler()
    LOGGER.cons_handler.setLevel(LOGGER.DEBUG)





from baopig.pybao.objectutilities import PrefilledFunction
from baopig import *
from baopig.unit_tests.testerscene import TesterScene



class PresentationScene(Scene):

    def __init__(self, app):

        Scene.__init__(self, app)

        GridLayer(self)

        Text(self, text="", row=len(self.default_layer))
        Text(self, text="Welcome to baopig version {}".format(version), row=len(self.default_layer))
        Text(self, text="You can look for the tutorial or experiment unit tests", row=len(self.default_layer))
        Text(self, text="", row=len(self.default_layer))

        Button(self, text="Unit Tests", command=PrefilledFunction(app.open, "UTMenu_Scene"),
               row=len(self.default_layer))
        Button(self, text="Tutorial", row=len(self.default_layer))

        UTMenu_Scene(app)


class UTMenu_Scene(Scene):

    def __init__(self, app):

        Scene.__init__(self, app)

        # l = MenuLayer(self)  # TODO
        GridLayer(self)

        Text(self, text="", row=0)
        Text(self, text="Which class do you want to test ?", row=1)
        Text(self, text="", row=2)

        back = Button(self, "Menu", command=PrefilledFunction(app.open, "PresentationScene"), col=1)
        back.bottomright = self.bottomright

        def get_ut_filenames():
            import os
            dir = os.path.dirname(os.path.realpath(__file__))[:-7] + "unit_tests"
            for root, dirs, files in os.walk(dir):
                for filename in files:
                    if filename.endswith(".py") and filename.startswith("ut_"):
                        yield filename[:-3]  # discard '.py'

        import importlib
        for filename in get_ut_filenames():
            print(filename)
            if filename in ("testerscene", "ut_baopig"):
                continue
            ut_file = importlib.import_module("baopig.unit_tests." + filename)
            # exec("from {} import get_ut_scenes".format(path))
            # from path import get_ut_scenes
            for zone in ut_file.ut_scenes:  # TODO : replace ut_scenes by ut_zones
                scene = TesterScene(app, zone)
                Button(self, row=len(self.default_layer), text=filename[3:],  # discard 'ut_'
                       command=PrefilledFunction(app.open, scene.name))




import pygame
from math import inf as math_inf
from baopig.pybao.issomething import *
from baopig.pybao.objectutilities import Object, TypedList
from baopig.ressources import *
from baopig.font.font import Font
from baopig.io import mouse
from baopig._lib import *


def set_indicator(self, text=None, get_text=None, indicator=None):
    """Create a text above the widget when hovered"""

    if self._indicator is not None:
        raise PermissionError("Can only have one indicator")
    if indicator is not None:
        self._indicator = indicator
    elif (text is not None) == (get_text is not None):
        raise PermissionError("You must only define one of text and get_text")
    elif text is None:
        self._indicator = DynamicText(
            self.parent, get_text,
            color=(255, 255, 255), font_height=15,
            pos=(0, -5), pos_location="bottom", pos_ref=self, pos_ref_location="top",
            background_color=(0, 0, 0, 192), padding=(8, 4), touchable=False,
        )
    else:
        self._indicator = Text(
            self.parent, text,
            color=(255, 255, 255), font_height=15,
            pos=(0, -5), pos_location="bottom", pos_ref=self, pos_ref_location="top",
            background_color=(0, 0, 0, 192), padding=(8, 4), touchable=False,
        )
    self._indicator.origin.config(from_hitbox=True)
    if self.layer.accept(self._indicator):
        self._indicator.swap_layer(self.layer)
        self._indicator.move_in_front_of(self)
    self._indicator.connect("wake", self.signal.HOVER)
    self._indicator.connect("asleep", self.signal.UNHOVER)
    if not self.is_hovered:
        self._indicator.asleep()
Hoverable.set_indicator = set_indicator


class _Line(ResizableWidget):
    """
    A Line is a component who only have text on its surface
    It have a transparent background
    It have an end string who is the separator between this line and the next one

    The aspects who might evolve :
        - the text
        - the font size
        - the font color
        - the end character

    When the hitbox size change, there is only one point who is used to locate the LineText : it is
    the origin. Exemple : if the origin is TOPLEFT, and the hitbox width grows, it will expend on the
    right side, because the origin is on the left. Ther is 9 possible positions for the origin :
        - topleft       - midtop        - topright
        - midleft       - center        - midright
        - bottomleft    - midbottom     - bottomright
    The default origin position is TOPLEFT


    Here is an example of font size :

        font_size = 63
        ____   _
        ____  |_   _   _         _             | 13 px  |
        ____  |   |_| |     |_| |_| |_|        | 37 px  | 63 px
        ____                 _|                | 13 px  |

    A Line is an element of a Paragraph, wich ends with a '\n'.

    """

    def __init__(self, parent, text, line_index):

        assert isinstance(parent, Text)

        self._line_index = line_index
        ResizableWidget.__init__(self,
            parent=parent,
            surface=pygame.Surface((parent.w, parent.font.height), pygame.SRCALPHA),
            layer=parent.lines,
            name="{}({})".format(self.__class__.__name__[1:], text),
        )

        self.__text = ""
        self.__end = ''
        self._text_with_end = None
        self._real_text = None
        # char_pos[i] est la distance entre left et la fin du i-eme caractere
        # Exemple : soit self.text = "Hello world"
        #           char_pos[6] = margin + distance entre le debut de "H" et la fin de "w"
        self._chars_pos = []

        self.config(text=text, end='\n', called_by_constructor=True)

    def __repr__(self):
        return "{}(index={}, text={})".format(self.__class__.__name__, self.line_index, self.text)

    def __str__(self):
        return self.text

    def _set_end(self, end):
        self.__end = end
        self._text_with_end = self.text + end
        self._real_text = self.text + ('' if end == '\v' else end)
    _end = property(lambda self: self.__end, _set_end)
    end = property(lambda self: self.__end)
    font = property(lambda self: self._parent._font)
    line_index = property(lambda self: self._line_index)
    real_text = property(lambda self: self._real_text)
    def _set_text(self, text):
        self.__text = text
        self._text_with_end = text + self.end
        self._real_text = text + ('' if self.end == '\v' else self.end)
    _text = property(lambda self: self.__text, _set_text)
    text = property(lambda self: self.__text)
    text_with_end = property(lambda self: self._text_with_end)

    def find_index(self, x, only_left=False, end_of_word=False):
        """
        Renvoie l'index correspondant a la separation de deux lettres la plus proche de x

        Example :
            x = 23              (23eme pixel a droite de self.left)
            find_index(x) -> 3  (position entre la 3eme et la 4eme lettre)

        Si only_left est demande, renvoie un index qui correspond a une separation a droite
        de x
        Si end_of_word est a True, l'index renvoye correspondra a une fin de mot
        Un mot est precede et suivi d'espaces ou d'un tiret
        Example:
            'Comment allez-vous ?' est forme des mots 'Comment', 'allez-', 'vous' et '?'
        """

        def ecart(x1, x2):
            return abs(x1 - x2)

        dist_from_closest_char = math_inf
        index_of_closest_char = None

        for index, char_pos in enumerate(self._chars_pos):

            if only_left and char_pos > x:
                break
            if ecart(x, char_pos) > dist_from_closest_char:
                break

            if end_of_word:
                if index != 0 and index != len(self.text):
                    def is_end_of_word(i):
                        if self.text[i] == ' ':
                            return True
                        if self.text[i-1] == '-':
                            return True
                        return False
                    if not is_end_of_word(index):
                        continue

            dist_from_closest_char = ecart(x, char_pos)
            index_of_closest_char = index

        return index_of_closest_char

    def find_mouse_index(self):
        """
        Return the closest index from mouse.x
        """

        return self.find_index(mouse.get_pos_relative_to(self)[0])

    def find_pixel(self, index):
        """
        Renvoi la distance entre hitbox.left et la fin du index-eme caractere
        """
        return self._chars_pos[index]

    def get_first_line_of_paragraph(self):
        if self.line_index == 0:
            return self
        i = self.line_index
        while self.parent.lines[i-1].end != '\n':
            i -= 1
        return self.parent.lines[i]

    def get_last_line_of_paragraph(self):
        i = self.line_index
        while self.parent.lines[i].end != '\n':
            i += 1
        return self.parent.lines[i]

    def get_paragraph(self):
        line = self.get_first_line_of_paragraph()
        while line.end != '\n':
            yield line
            line = self.parent.lines[line.line_index+1]
        yield line

    def get_paragraph_text(self):
        return ''.join(line.real_text for line in self.get_paragraph())[:-1]  # discard '\n'

    def get_paragraph_text_with_end(self):
        return ''.join(line.text_with_end for line in self.get_paragraph())

    def insert(self, char_index, string):
        """
        Insert a string inside a line (not after the end delimitation)
        """

        if string:
            self.config(text=self.text[:char_index] + string + self.text[char_index:])

    def pop(self, index):
        """
        Remove one character from line.real_text
        """

        if index < 0:
            index = len(self.real_text) + index
        if self.end != '\v' and index == len(self.real_text)-1:
            if self.line_index == len(self.parent.lines)-1:
                return  # pop of end of text
            self.config(end='\v')
        else:
            self.config(text=self.text[:index] + self.text[index+1:])

    def config(self, text=None, end=None, called_by_constructor=False):

        # if not called_by_constructor:

        with paint_lock:

            if end is not None:
                assert end in ('\n', ' ', '\v')
                if self.parent.max_width is None:
                    assert end == '\n'
                self._end = end
            if text is not None:
                if '\t' in text:
                    text = str.replace(text, '\t', '    ')  # We can also replace it at rendering
                if '\v' in text:
                    text = str.replace(text, '\v', '')
                assert isinstance(text, str)
                self._text = text

            if self.end != '\n' and self.line_index == len(self.parent.lines)-1:
                raise PermissionError

            if called_by_constructor is False and len(tuple(self.get_paragraph())) > 0:

                self._text = self.get_paragraph_text()

                for line in tuple(self.get_paragraph()):
                    if line != self:
                        line.kill()
                self.parent._pack()

                self._end = '\n'

            if '\n' in self.text:
                self.__class__(
                    parent=self.parent,
                    text=self.text[self.text.index('\n')+1:],
                    line_index=self.line_index + .5
                )
                self._text = self.text[:self.text.index('\n')]

            self.update_char_pos()
            if self.parent.max_width is not None and self.font.get_width(self.text) > self.parent.max_width:

                max_width = self.parent.max_width
                assert self.find_pixel(1) <= max_width, "The max_width is too little : " + str(max_width)

                if True:
                    index_end = index_newline_start = self.find_index(
                        max_width, only_left=True, end_of_word=True)
                    if index_end == 0:
                        sep = '\v'
                        index_newline_start = index_end = self.find_index(
                            max_width, only_left=True, end_of_word=False)
                    else:
                        end = self.text[index_end]
                        sep = ' ' if end == ' ' else '\v'  # else, the word is of type 'smth-'
                        if end == ' ': index_newline_start += 1
                    LineClass = _SelectableLine if isinstance(self, Selectable) else _Line
                    self.__class__(
                        parent=self.parent,
                        text=self.text[index_newline_start:],
                        line_index=self.line_index + .5,  # the line will correct itself
                    )
                    self._end = sep
                    self._text = self.text[0:index_end]
                    self.update_char_pos()

            font_render = self.font.render(self.text)
            surf_w = font_render.get_width()
            # if self.parent.max_width is not None: surf_w = max(self.parent.w, surf_w)
            surface = pygame.Surface((surf_w, self.font.height), pygame.SRCALPHA)
            surface.blit(font_render, (0, 0))
            """if self.parent.max_width is None:
                surface.blit(font_render, (0, 0))
            elif self.parent.align_mode == "left":
                surface.blit(font_render, (0, 0))
            elif self.parent.align_mode == "center":
                surface.blit(font_render, (int((surf_w - font_render.get_width()) / 2), 0))
            elif self.parent.align_mode == "right":
                surface.blit(font_render, (surf_w - font_render.get_width(), 0))
            else:
                raise ValueError"""
            self.set_surface(surface)
            self.parent._pack()

    def update_char_pos(self):
        """
        Actualise les valeurs de char_pos

        Appele lors de la creation de LineSelection et lorsque le texte change

        char_pos[i] est la distance entre hitbox.left et la fin du i-eme caractere
        Exemple :
                    Soit self.text = "Hello world"
                    char_pos[6] = margin + distance entre le debut de "H" et la fin de "w"
        """
        self._chars_pos = [0]
        text = ''
        for char in self.text:
            text += char
            self._chars_pos.append(self.font.get_width(text))


class _SelectableLine(_Line, Selectable):
    """
    You are selecting a SelectableLine when :
        - A condition described in Selectable is verified
        - A cursor moves while Maj key is pressed
    """
    
    def __init__(self, *args, **kwargs):

        _Line.__init__(self, *args, **kwargs)
        Selectable.__init__(self)

        def connect():
            self.connect("handle_selector_link", self.selector.signal.LINK)
        self.parent.send_request(connect)

        self._selection_ref = lambda: None

    selection = property(lambda self: self._selection_ref())

    def select(self):

        selection = self.selector.selection_rect
        assert self.abs_hitbox.colliderect(selection.abs_hitbox)
        if self.selection is None and not selection.w and not selection.h:
            return

        if self.selection is None:
            _LineSelection(self)

        selecting_line_end = False
        if self.abs.top <= selection.start[1] < self.abs.bottom:
            start = self.find_index(selection.start[0] - self.abs.left)
        elif selection.start[1] < self.abs.top: start = 0
        else:
            start = len(self.text)
            if self is not self.parent.lines[-1]: selecting_line_end = True
        if self.abs.top <= selection.end[1] < self.abs.bottom:
            end = self.find_index(selection.end[0] - self.abs.left)
        elif selection.end[1] < self.abs.top: end = 0
        else:
            end = len(self.text)
            if self is not self.parent.lines[-1]: selecting_line_end = True

        start, end = sorted((start, end))
        self.selection.config(start, end, selecting_line_end)

    def get_selected_data(self):
        if self.selection is None:
            return ''
        return self.selection.get_data()

    def handle_selector_link(self):

        if mouse.has_triple_clicked:
            if self.parent.collidemouse() and self.hitbox.top <= mouse.get_pos_relative_to(self.parent)[1] < self.hitbox.bottom:
                with paint_lock:
                    self.selector.close_selection()
                    self.selector.start_selection((self.abs.left, self.abs.top))
                    self.selector.end_selection((self.abs.right, self.abs.top), visible=False)  # not + 1 so we don't see the selection rect
                # print("selection :", self.selection.index_start, self.selection.index_end)
        elif mouse.has_double_clicked:
            if self.parent.collidemouse() and self.hitbox.top <= mouse.get_pos_relative_to(self.parent)[1] < self.hitbox.bottom:
                self.select_word(self.find_mouse_index())
                # print("selection :", self.selection.index_start, self.selection.index_end)

    def select_word(self, index):
        """
        Selectionne le mot le plus proche de index
        """

        separators = " ,„.…:;/\'\"`´”’" \
                     "=≈≠+-±–*%‰÷∞√∫" \
                     "()[]{}<>≤≥«»" \
                     "?¿!¡@©®ª#§&°" \
                     "‹◊†¬•¶|^¨~" \
                     ""

        index_start = index_end = index
        while index_start > 0 and self.text[index_start - 1] not in separators:
            index_start -= 1
        while index_end < len(self.text) and self.text[index_end] not in separators:
            index_end += 1

        # Si index n'est pas sur un mot, on selectionne tout le label
        if index_start == index_end == index:
            index_start = 0
            index_end = len(self.text)

        with paint_lock:
            if self.selector.is_selecting:
                self.selector.close_selection()
            if index_start == 0:
                self.selector.start_selection((self.abs.left, self.abs.top))
            else:
                self.selector.start_selection((self.abs.left + self.find_pixel(index_start), self.abs.top))
            if index_end == len(self.text):
                self.selector.end_selection((self.abs.right, self.abs.top), visible=False)
            else:
                self.selector.end_selection((self.abs.left + self.find_pixel(index_end), self.abs.top), visible=False)

    def unselect(self):
        if self.selection is not None:
            self.selection.kill()


class _LineSelection(Rectangle):
    """
    A LineSelection is a Rectangle with a light blue color : (167, 213, 255, 127)
    Each Line can have a LineSelection

    When you click on a SelectableLine, and then move the mouse while its pressed,
    you are selecting the SelectableLine
    The size and the position of the LineSelection object change according to your mouse

    When you double-click on a SelectableLine, it select a word
    When you triple-click on a SelectableLine, it select the whole line text
    """

    STYLE = Rectangle.STYLE.substyle()
    STYLE.modify(
        color = "theme-color-selection",
        border_width = 0
    )

    def __init__(self, line):

        assert isinstance(line, _SelectableLine)

        self._line_index = line.line_index
        Rectangle.__init__(self,
            parent=line.parent,
            pos=line.topleft,
            size=(0, line.h),
            name=line.name+" -> selection"
        )

        # self.is_selecting = False  # True if the user is pressing the mouse button for a selection
        self._index_start = self._index_end = 0
        self._is_selecting_line_end = False
        self._line_ref = line.get_weakref()

        # Initializations
        # self.move_behind(self.line)
        self.line._selection_ref = self.get_weakref()
        self.depend_on(self.line)
        self.swap_layer("line_selections")

    index_end = property(lambda self: self._index_end)
    index_start = property(lambda self: self._index_start)
    line = property(lambda self: self._line_ref())
    line_index = property(lambda self: self._line_index)
    text = property(lambda self: self.line._text)

    def config(self, index_start, index_end, selecting_line_end):

        index_start, index_end = sorted((index_start, index_end))
        self.set_start(index_start)
        self.set_end(index_end, selecting_line_end)
        # TODO : remove set_start and set_end

    def get_data(self):
        end = self.line.end if self._is_selecting_line_end else ''
        if end == '\v': end = ''
        return self.line.text[self.index_start:self.index_end] + end

    def set_end(self, index, selecting_line_end):

        self._index_end = index
        self._is_selecting_line_end = selecting_line_end
        if selecting_line_end:
            assert self.line is not self.parent.lines[-1]

        if self._is_selecting_line_end:
            self.resize_width(self.line.width -
                              self.line.find_pixel(self.index_start))
        else:
            self.resize_width(abs(self.line.find_pixel(self.index_end) -
                                  self.line.find_pixel(self.index_start)))
        self.left = self.line.x + self.line.find_pixel(self.index_start)

    def set_start(self, index):

        if index == self.index_start:
            return

        self._index_start = self._index_end = index
        self.resize_width(0)
        self.left = self.line.find_pixel(self._index_start)

        if self.is_sleeping:
            self.wake()
        self.show()


# TODO : Line (text with only one line)
class Text(Zone):

    STYLE = Zone.STYLE.substyle()
    STYLE.modify(
        width=0,
        height=0,
    )
    STYLE.create(
        font_file = None,
        font_height = 15,
        color = "theme-color-font",
        bold = False,
        italic = False,
        underline = False,
        align_mode = "left",
        max_width=None,
    )
    STYLE.set_type("font_height", int)
    STYLE.set_type("color", Color)
    STYLE.set_type("bold", bool)
    STYLE.set_type("italic", bool)
    STYLE.set_type("underline", bool)
    STYLE.set_type("align_mode", str)
    STYLE.set_type("padding", MarginType)
    STYLE.set_constraint("font_height", lambda val: val > 0, "a text must have a positive font height")
    STYLE.set_constraint("font_file", lambda val: (val is None) or isinstance(val, str), "must be None or a string")
    STYLE.set_constraint("align_mode", lambda val: val in ("left", "center", "right"), "must be 'left', 'center' or 'right'")

    # TODO : align modes
    # LEFT_MODE = 50
    # CENTER_MODE = 51
    # RIGHT_MODE = 52

    def __init__(self,
        parent,
        text=None,
        font=None,  # TODO : remove it
        font_file=None,
        font_height=None,
        color=None,
        bold=False,
        italic=False,
        underline=False,
        max_width=None,
        padding=None,
        selectable=True,
        **options
    ):

        # if mode is None: mode = Text.LEFT_MODE

        if max_width is not None:
            options["max_width"] = max_width
            assert isinstance(max_width, int)
            assert max_width > 0
        assert isinstance(selectable, bool)

        Zone.__init__(self, parent, **options)

        if font is None:
            self.inherit_style(
                parent,
                font_file=font_file,
                font_height=font_height,
                color=color,
                bold=bold,
                italic=italic,
                underline=underline,
                padding=padding,
            )
            font = Font(
                file=self.style["font_file"],
                height=self.style["font_height"],
                color=self.style["color"],
                bold=self.style["bold"],
                italic=self.style["italic"],
                underline=self.style["underline"],
                text_owner=self,
            )
        else:
            assert isinstance(font, Font), str(font)
            font = font.copy(text_owner=self)

        self._font = font
        self._max_width = self.style["max_width"]
        self._min_width = self.font.get_width("m")
        self._is_selectable = selectable
        self._lines_pos = []
        self._align_mode = self.style["align_mode"]
        self._padding = self.style["padding"]
        self.has_locked.text = False

        if self.max_width is not None:
            self.resize_width(self.max_width)
            self.lock_width(True)

        self.line_selections = Layer(self, _LineSelection, name="line_selections", touchable=False, sort_by_pos=True)
        # self.lines = GridLayer(self, "lines", _Line)
        self.lines = Layer(self, _Line, name="lines", default_sortkey=lambda line: line.line_index)
        # self.layers.add_layer("line_selections", CompsClass=_LineSelection, touchable=False)
        # self.layers.add_layer("lines", CompsClass=_Line, default_sortkey=lambda line: line.line_index)
        # self.line_selections.move_behind("lines")
        self.set_text(text)

    align_mode = property(lambda self: self._align_mode)
    font = property(lambda self: self._font)
    is_selectable = property(lambda self: self._is_selectable)
    max_width = property(lambda self: self._max_width)
    padding = property(lambda self: self._padding)

    def _pack(self):

        if self.align_mode == "center":  # only usefull for the widget creation
            if self.max_width is None:
                centerx = int(max(l.w for l in self.lines) / 2)
            else:
                centerx = int(self.max_width / 2)

        self.lines.sort()
        h = self.padding.top
        for i, line in enumerate(self.lines):
            line.topleft = (self.padding.left, h)
            line._line_index = i
            h = line.bottom

            if self.align_mode == "left":
                line.left = self.padding.left
            elif self.align_mode == "center":
                line.centerx = centerx
            elif self.align_mode == "right":
                line.right = self.width - self.padding.right

        self.adapt(self.lines, padding=self.padding, horizontally=self.max_width is None)

        self._lines_pos = []
        for line in self.lines:
            self._lines_pos.append(line.top)

    def _find_index(self, pos):
        """
        Renvoie l'index correspondant a l'espace entre deux caracteres le plus proche

        Exemple :
            pos = (40, 23)                             (40eme pixel a partir de self.left, 23eme pixel sous self.top)
            find_index(pos) -> self.find_index(2, 13)  (pos est sur le 14eme caractere de la 3eme ligne)
        """

        if pos[1] < 0:
            return self.lines[0].find_index(pos[0])
        elif pos[1] >= self.h:
            return self.find_index(len(self.lines) - 1, self.lines[-1].find_index(pos[0]))
        else:
            for line_index, line in enumerate(self.lines):
                if pos[1] < line.bottom:
                    return self.find_index(line_index, line.find_index(pos[0]))
        assert self.lines[-1].bottom == self.h, str(self.lines[-1].bottom) + ' ' + str(self.h)
        raise Exception

    def find_index(self, line_index, char_index=None):
        """
        This method return the total index from a line index and a character index

        Example:
            text = "Hello\n"
                   "world"
            text.find_index(1, 2) -> index between 'o' and 'r'
                                  -> 8

        WARNING : this method result don't always match with text.index('r'), when
                  the text is cut inside a word or after a '-', we need two different
                  indexes for the end of the line and the start of the next line
        """
        if char_index is None:
            return self._find_index(pos=line_index)

        text_index = 0
        for i, line in enumerate(self.lines):
            if i == line_index:
                break
            text_index += len(line.real_text)
        return text_index + char_index

    def _find_indexes(self, pos):
        """
        Renvoie l'index correspondant a l'espace entre deux caracteres le plus proche

        Exemple :
            pos = (40, 23)                             (40eme pixel a partir de self.left, 23eme pixel sous self.top)
            find_index(pos) -> self.find_index(2, 13)  (pos est sur le 14eme caractere de la 3eme ligne)
        """

        if pos[1] < 0:
            return 0, 0
        elif pos[1] >= self.h:
            return len(self.lines) - 1, len(self.lines[-1].text)
        else:
            for line_index, line in enumerate(self.lines):
                if line.bottom > pos[1]:
                    return line_index, line.find_index(pos[0])

        raise Exception

    def _find_indexes_corrected(self, line_index, char_index):

        text_index = 0
        for i, line in enumerate(self.lines):
            if i == line_index:
                break
            text_index += len(line.real_text)
        text_index += char_index
        if text_index < 0: return 0, 0, 0
        if text_index > len(self.text):
            return len(self.text), len(self.lines)-1, len(self.lines[-1].text)

        if not 0 <= char_index <= len(self.lines[line_index].text):
            char_index = text_index
            for line_index, line in enumerate(self.lines):
                if char_index <= len(line.text):
                    assert 0 <= char_index <= len(line.text), char_index
                    break
                char_index -= len(line.real_text)

        assert text_index == self.find_index(line_index, char_index), "{}, {}, {}".format(text_index, line_index, char_index)

        return text_index, line_index, char_index

    def find_indexes(self, text_index=None, line_index=None, char_index=None, pos=None):
        """
        Renvoie l'inverse de self.find_index(line_index, char_index)

        Example:
            text = "Hello\n"
                   "world"
            text.find_indexes(8) -> (1, 2)
        """

        if line_index is not None:
            assert text_index is None
            assert char_index is not None
            assert pos is None
            return self._find_indexes_corrected(line_index=line_index, char_index=char_index)

        if pos is not None:
            assert text_index is None
            assert line_index is None
            assert char_index is None
            return self._find_indexes(pos=pos)

        if text_index < 0: return 0, 0
        char_index = text_index
        for line_index, line in enumerate(self.lines):
            if char_index <= len(line.text):
                assert 0 <= char_index <= len(line.text), char_index
                return line_index, char_index
            char_index -= len(line.real_text)

        # The given text_index is too high
        return line_index, len(line.text)

    def find_mouse_index(self):
        """
        Return the closest index from mouse.x
        """
        return self.find_index(mouse.get_pos_relative_to(self))

    def find_mouse_indexes(self):
        """
        Return the closest indexes from mouse.x
        """
        return self.find_indexes(pos=mouse.get_pos_relative_to(self))

    def _find_pos(self, text_index):

        line_index, char_index = self.find_indexes(text_index=text_index)
        return self.find_pos(line_index=line_index, char_index=char_index)

    def find_pos(self, text_index=None, line_index=None, char_index=None):
        """
        Effectue l'inverse de find_indexes
        La position renvoyee est relative a self.topleft
        """
        if text_index is not None: return self._find_pos(text_index=text_index)
        return self.lines[line_index].find_pixel(char_index), self._lines_pos[line_index]

    def get_selected_data(self):
        if self.is_selectable:
            return ''.join(line.get_selected_data() for line in self.lines)

    def get_text(self):
        return ''.join(line.real_text for line in self.lines)[:-1]  # Discard last \n
    text = property(get_text)

    def lock_text(self, locked=True):

        self.has_locked.text = locked

    def set_max_width(self, max_width):

        assert isinstance(max_width, int)
        assert max_width > 0
        if self._min_width > max_width:
            raise PermissionError(f"The max_width is too little : {max_width}")
        self._max_width = max_width
        self.lock_width(False)
        self.resize_width(self.max_width)
        self.lock_width(True)
        self.set_text(self.get_text())

    def set_text(self, text):

        if self.has_locked.text: return
        with paint_lock:

            first_line = self.lines[0] if self.lines else None
            for child in tuple(self.lines):
                assert child in self.children
                assert self == child.parent
                if child != first_line:
                    child.kill()
            try:
                assert len(self.lines) in (0, 1), self.lines
            except Exception as e:
                raise e

            if first_line is not None:
                first_line.config(text=text, end='\n')
            else:
                LineClass = _SelectableLine if self.is_selectable else _Line
                LineClass(
                    parent=self,
                    text=text,
                    line_index=0,
                )
            self._pack()
            self._name = self.lines[0].text


class DynamicText(Text, Runable):

    def __init__(
        self,
        parent,
        get_text,
        **kwargs
    ):

        assert callable(get_text), get_text

        Text.__init__(
            self,
            parent=parent,
            text=str(get_text()),
            **kwargs
        )
        Runable.__init__(self)

        self.get_new_text = get_text

        self.start_running()

    def run(self):

        new_text = str(self.get_new_text())
        if new_text != self.text:
            self.set_text(new_text)


class TextLabel(Text):
    """
    A TextLabel is a Label who only contains text. The interest of TextLabel is to restrict
    the Text sizes.

    If a width is given, then the Text won't be longer, just like Text.max_width
    If a height is given, same thing, the Text won't be higher.

    When a text is longer than the required width, there is 3 solutions :
        - The text inserts an end of line (solution opered by Text.max_width, not possible in TextLabel)
        - The text font height is reduced
        - The text stay longer, a scroll could allow to read it
    When a text is higher than the required height, there is 2 solutions :
        - The text font height is reduced
        - The text stay higher, a scroll could allow to read it


    """

    def __init__(self, parent, *args, width=None, height=None, scrollable=False, **options):

        assert "max_width" not in options, "Use 'width' instead"
        Text.__init__(self, parent, *args, **options)

        self.set_size(width, height)

    def set_size(self, width=None, height=None):

        if (width is not None) or height is not None:
            if not self.padding.is_null:
                if self.align_mode != "center":
                    print(self.padding)
                    raise PermissionError("Cannot define padding and size on a TextLabel")
            if width is not None:
                plus = (width - self.width) / 2
                self.padding.left += plus
                self.padding.right += plus
            if height is not None:
                plus = (height - self.height) / 2
                self.padding.top += plus
                self.padding.bottom += plus
            self._pack()







import pygame
from baopig.pybao.objectutilities import Object
from baopig.time.timer import RepeatingTimer
from .logging import LOGGER
from .clipboard import clipboard


class KeyEvent(Object):

    def __init__(self, event):

        Object.__init__(self, type=event.type, **event.__dict__)
        keyboard.last_event = self

    def __str__(self):
        global __key_events
        return "<KeyEvent({}-{} {})>".format(
            self.type,
            "KEYDOWN" if self.type == keyboard.KEYDOWN else "KEYUP" if self.type == keyboard.KEYUP else "Unknown",
            self.__dict__
        )


class _Keyboard:

    # copied from pygame.constants
    if True:
        KEYDOWN = pygame.KEYDOWN
        KEYUP = pygame.KEYUP

        MOD_ALT = 768
        MOD_CAPS = 8192
        MOD_CTRL = 192
        MOD_LALT = 256
        MOD_LCTRL = 64
        MOD_LMETA = 1024
        MOD_LSHIFT = 1
        MOD_META = 3072
        MOD_MODE = 16384
        MOD_NONE = 0
        MOD_NUM = 4096
        MOD_RALT = 512
        MOD_RCTRL = 128
        MOD_RMETA = 2048
        MOD_RSHIFT = 2
        MOD_SHIFT = 3

        _0 = 48
        _1 = 49
        _2 = 50
        _3 = 51
        _4 = 52
        _5 = 53
        _6 = 54
        _7 = 55
        _8 = 56
        _9 = 57
        a = 97
        AMPERSAND = 38
        ASTERISK = 42
        AT = 64
        b = 98
        BACKQUOTE = 96
        BACKSLASH = 92
        BACKSPACE = 8
        BREAK = 318
        c = 99
        CAPSLOCK = 301
        CARET = 94
        CLEAR = 12
        COLON = 58
        COMMA = 44
        d = 100
        DELETE = 127
        DOLLAR = 36
        DOWN = 274
        e = 101
        END = 279
        EQUALS = 61
        ESCAPE = 27
        EURO = 321
        EXCLAIM = 33
        f = 102
        F1 = 282
        F10 = 291
        F11 = 292
        F12 = 293
        F13 = 294
        F14 = 295
        F15 = 296
        F2 = 283
        F3 = 284
        F4 = 285
        F5 = 286
        F6 = 287
        F7 = 288
        F8 = 289
        F9 = 290
        FIRST = 0
        g = 103
        GREATER = 62
        h = 104
        HASH = 35
        HELP = 315
        HOME = 278
        i = 105
        INSERT = 277
        j = 106
        k = 107
        KP0 = 256
        KP1 = 257
        KP2 = 258
        KP3 = 259
        KP4 = 260
        KP5 = 261
        KP6 = 262
        KP7 = 263
        KP8 = 264
        KP9 = 265

        KP_DIVIDE = 267
        KP_ENTER = 271
        KP_EQUALS = 272
        KP_MINUS = 269
        KP_MULTIPLY = 268
        KP_PERIOD = 266
        KP_PLUS = 270

        l = 108
        LALT = 308
        LAST = 323
        LCTRL = 306
        LEFT = 276
        LEFTBRACKET = 91
        LEFTPAREN = 40
        LESS = 60
        LMETA = 310
        LSHIFT = 304
        LSUPER = 311
        m = 109
        MENU = 319
        MINUS = 45
        MODE = 313
        n = 110
        NUMLOCK = 300
        o = 111
        p = 112
        PAGEDOWN = 281
        PAGEUP = 280
        PAUSE = 19
        PERIOD = 46
        PLUS = 43
        POWER = 320
        PRINT = 316
        q = 113
        QUESTION = 63
        QUOTE = 39
        QUOTEDBL = 34
        r = 114
        RALT = 307
        RCTRL = 305
        RETURN = 13
        RIGHT = 275
        RIGHTBRACKET = 93
        RIGHTPAREN = 41
        RMETA = 309
        RSHIFT = 303
        RSUPER = 312
        s = 115
        SCROLLOCK = 302
        SEMICOLON = 59
        SLASH = 47
        SPACE = 32
        SYSREQ = 317
        t = 116
        TAB = 9
        u = 117
        UNDERSCORE = 95
        UNKNOWN = 0
        UP = 273
        v = 118
        w = 119
        x = 120
        y = 121
        z = 122

    def __init__(self):

        # self._keys = [0] * 512          # 512 = len(pygame.key.get_pressed())
        self._keys = {}
        self._application = None
        class Mod:  # todo : update
            def __init__(self):
                self.l_alt = False
                self.r_alt = False
                self.alt = self.l_alt or self.r_alt
                self.l_cmd = False
                self.r_cmd = False
                self.cmd = self.l_cmd or self.r_cmd
                self.l_ctrl = False
                self.r_ctrl = False
                self.ctrl = self.l_alt or self.r_ctrl
                self.l_maj = False
                self.r_maj = False
                self.maj = self.l_maj or self.r_maj
            def __str__(self):
                return Object.__str__(self)
        self._mod = Mod()
        self.last_event = None

        # repeat
        self._keys_time = [None] * 323  # the time where the key have been pressed
        self._is_repeating = False
        self._repeat_first_delay = None
        self._repeat_delay = None

    application = property(lambda self: self._application)
    is_repeating = property(lambda self: self._is_repeating)
    mod = property(lambda self: self._mod)

    def _release_all(self):

        for key in tuple(self._keys):
            self.receive(Object(type=pygame.KEYUP, key=key))

    def is_pressed(self, key):
        """Return True if the key with identifier 'key' (an integer) is pressed"""

        # You can write bp.keyboard.is_pressed("z")
        if isinstance(key, str):
            key = getattr(self, key)

        try:
            return bool(self._keys[key])
        except KeyError:
            # Here, the key has never been pressed
            return 0

    def receive(self, event):
        """Receive pygame events from the application"""

        # ACTUALIZING KEYBOARD STATES
        if event.type == pygame.KEYDOWN:
            self._keys[event.key] = 1
            if self._is_repeating and self._keys_time[event.key] is None:
                repeat = RepeatingTimer((self._repeat_first_delay / 1000, self._repeat_delay / 1000),
                                        pygame.event.post, event)
                repeat.start()
                self._keys_time[event.key] = repeat
        elif event.type == pygame.KEYUP:
            if self._keys[event.key] == 0:
                return  # The KEYDOWN have been skipped, so we skip the KEYUP
            self._keys[event.key] = 0
            if self._is_repeating:
                assert self._keys_time[event.key] is not None
                self._keys_time[event.key].cancel()
                self._keys_time[event.key] = None
        else:
            LOGGER.warning("Unexpected event : {}".format(event))
            return
        KeyEvent(event)  # keyboard.last_event

        if event.key == pygame.K_RALT:
            self.mod.r_alt = event.type == pygame.KEYDOWN
            self.mod.alt = self.mod.l_alt or self.mod.r_alt
        elif event.key == pygame.K_LALT:
            self.mod.l_alt = event.type == pygame.KEYDOWN
            self.mod.alt = self.mod.l_alt or self.mod.r_alt
        elif event.key == pygame.K_RMETA:
            self.mod.r_cmd = event.type == pygame.KEYDOWN
            self.mod.cmd = self.mod.l_cmd or self.mod.r_cmd
        elif event.key == pygame.K_LMETA:
            self.mod.l_cmd = event.type == pygame.KEYDOWN
            self.mod.cmd = self.mod.l_cmd or self.mod.r_cmd
        elif event.key == pygame.K_RCTRL:
            self.mod.r_ctrl = event.type == pygame.KEYDOWN
            self.mod.ctrl = self.mod.l_ctrl or self.mod.r_ctrl
        elif event.key == pygame.K_LCTRL:
            self.mod.l_ctrl = event.type == pygame.KEYDOWN
            self.mod.ctrl = self.mod.l_ctrl or self.mod.r_ctrl
        elif event.key == pygame.K_RSHIFT:
            self.mod.r_maj = event.type == pygame.KEYDOWN
            self.mod.maj = self.mod.l_maj or self.mod.r_maj
        elif event.key == pygame.K_LSHIFT:
            self.mod.l_maj = event.type == pygame.KEYDOWN
            self.mod.maj = self.mod.l_maj or self.mod.r_maj

    def set_repeat(self, first_delay, delay):
        """Control how held keys are repeated, with delays in milliseconds"""
        # This solve a bug in pygame, who can't repeat two keys

        assert first_delay >= 0
        assert delay > 0

        if 1 in self._keys:
            raise PermissionError("You must set the keys repeat before launch the application")

        self._is_repeating = True
        self._repeat_first_delay = first_delay
        self._repeat_delay = delay


keyboard = _Keyboard()




import pygame
from .utilities import paint_lock
from .widget import Widget, _Origin


class _Origin2(_Origin):
    """Optimized origin, only its position is configurable"""
    def __init__(self, owner):
        self._asked_pos = owner.pos
        self._from_hitbox = False
        self._owner_ref = owner.get_weakref()
        self._location = "topleft"
        self._reference_ref = self.owner.parent.get_weakref()
        self._reference_location = "topleft"

    pos = property(lambda self: self._asked_pos)

    def _reset_asked_pos(self):
        """
        This method should only be called by Widget._move()
        If the asked_pos was including percentage, and the percentage don't match anymore, it will be
        replaced by an integer value
        """
        self._asked_pos = self.owner.pos

    @staticmethod
    def accept(coord):
        if hasattr(coord, "__iter__"):
            if len(coord) != 2: return False
            return False not in (_Origin.accept(c) for c in coord)
        return isinstance(coord, (int, float))

    def config(self, pos=None, location=None, reference_comp=None, reference_location=None,
               from_hitbox=None, locked=None):
        raise PermissionError("Nope")


class AnimatedWidget(Widget):
    """
    Optimized class for widgets who move a lot
    """
    def __init__(self, *args, **kwargs):
        Widget.__init__(self, *args, **kwargs)
        assert not isinstance(self.origin.pos[0], str)
        assert not isinstance(self.origin.pos[1], str)
        assert self.origin.location == "topleft"
        assert self.origin.reference == self.parent
        assert self.origin.reference_location == "topleft"
        assert self.window is None
        self.move_at(kwargs["pos"])
        self._origin = _Origin2(self)

    # abs = abs_rect =    property(lambda self: self._abs_rect)
    # auto = auto_rect =  property(lambda self: self._auto_rect)
    # window =            property(lambda self: self._window)
    # hitbox =            property(lambda self: self._hitbox)
    # abs_hitbox =        property(lambda self: self._abs_hitbox)
    # auto_hitbox =       property(lambda self: self._auto_hitbox)

    def _move(self, dx, dy):

        old_hitbox = tuple(self.hitbox)
        with paint_lock:

            self.origin._asked_pos = (self.rect.left+dx, self.rect.top+dy)
            pygame.Rect.__setattr__(self.rect, "topleft", self.origin._asked_pos)
            self._hitbox = self.rect
            pygame.Rect.__setattr__(self.abs_rect, "topleft", (self.parent.abs.left + self.left, self.parent.abs.top + self.top))
            self._abs_hitbox = self.abs_rect

            self.signal.MOTION.emit(dx, dy)
            if self.is_visible:
                self.parent._warn_change(self.hitbox.union(old_hitbox))

    def set_window(self, window, follow_movements=None):

        raise PermissionError("A window is over-consuming...")

    def _set_surface(self, surface):

        with paint_lock:
            self._surface = surface
            size = surface.get_size()
            assert size != self.size
            if size != self.size:
                pygame.Rect.__setattr__(self.rect, "size", size)
                pygame.Rect.__setattr__(self.abs_rect, "size", size)
                pygame.Rect.__setattr__(self.auto_rect, "size", size)
                pygame.Rect.__setattr__(self.hitbox, "size", size)
                pygame.Rect.__setattr__(self.abs_hitbox, "size", size)
                pygame.Rect.__setattr__(self.auto_hitbox, "size", size)






class Hoverable:
    """
    Abstract class for widgets who need to handle when they are hovered or unhovered
    A widget is hovered when the mouse is over it

    Example:

    class Obj(Widget, Hoverable)
        def __init__(self, blabla):
            Widget.__init__(self, blabla)
            Hoverable.__init__(self)
            self.signal.HOVER.add_command(lambda: print("Hovered"))
            self.signal.UNHOVER.add_command(lambda: print("Unhovered"))
    """

    def __init__(self):

        self._is_hovered = False
        self._indicator = None

        self.create_signal("HOVER")
        self.create_signal("UNHOVER")

        self.connect("handle_hover", self.signal.HOVER)
        self.connect("handle_unhover", self.signal.UNHOVER)

    indicator = property(lambda self: self._indicator)
    is_hovered = property(lambda self: self._is_hovered)

    def handle_hover(self):
        """Stuff to do when the widget gets hovered by mouse"""

    def handle_unhover(self):
        """Stuff to do when the widget is not hovered by mouse anymore"""

    def set_indicator(self, text=None, get_text=None, indicator=None):
        """
        Create a text above the widget when hovered

        :param text:
        :param get_text:
        :param indicator:
        """
        # NOTE : set_indicator function is added in baopig.widgets.Text






import functools
import pygame
from baopig.pybao.issomething import is_point
from baopig.pybao.objectutilities import PackedFunctions, PrefilledFunction, TypedDict, WeakTypedSet, Object
from baopig._debug import debug_with_assert
from baopig.ressources import *
from baopig.communicative import ApplicationExit
from baopig.io import LOGGER, mouse, keyboard

"""

    NOTE : handle functionnement
           In order to execute a function at handle_something, write :
                handle_something.add(function)
                
"""

import threading
paint_lock = threading.RLock()

# TODO : deproteger les attributs handle_something


class Size(tuple):

    def __init__(self, size):
        assert len(self) == 2, f"Wrong length : {self} (a size only have 2 elements)"
        for coord in self:
            assert isinstance(coord, (int, float)), f"Wrong value in size : {coord} (must be a number)"
            assert coord >= 0, f"Wrong value in size : {coord} (must be positive)"


class Color(pygame.Color):
    """
    Very close to pygame.Color :

    Color(tuple) -> Color
    Color(name) -> Color
    Color(r, g, b, a) -> Color
    Color(rgbvalue) -> Color
    pygame object for color representations
    """

    def __init__(self, *args, transparency=None, **kwargs):

        if transparency is None:
            try:
                pygame.Color.__init__(self, *args, **kwargs)
            except ValueError:
                pygame.Color.__init__(self, *args[0])

        else:
            try:
                color = pygame.Color(*args, **kwargs)
            except ValueError:
                color = pygame.Color(*args[0])
            pygame.Color.__init__(self, color.r, color.g, color.b, transparency)

    def set_hsv(self, val):
        if val[1] > 100:
            val = val[0], 100, val[2]
        try:
            self.hsva = val + (100,)
        except ValueError as e:
            raise ValueError(str(e) + f" : {val}")
    hsv = property(lambda self: self.hsva[:-1], set_hsv)

    def set_hsl(self, val):
        if val[1] > 100:
            val = val[0], 100, val[2]
        try:
            self.hsla = val + (100,)
        except ValueError as e:
            raise ValueError(str(e) + f" : {val}")
    hsl = property(lambda self: self.hsla[:-1], set_hsl)

    def set_hue(self, h):
        self.hsv = (h,) + self.hsv[1:]
    h = property(lambda self: self.hsva[0], set_hue)

    def get_saturation(self):
        s = self.hsva[1]
        if s > 100:  # sometime 100.00000000000001
            s = 100
        return s
    def set_saturation(self, s):
        self.hsv = self.h, s, self.v
    s = property(get_saturation, set_saturation)

    def set_value(self, v):
        self.hsv = self.hsva[:2] + (v,)
    v = property(lambda self: self.hsva[2], set_value)

    def set_lightness(self, l):
        self.hsl = self.hsla[:2] + (l,)
    l = property(lambda self: self.hsla[2], set_lightness)

    def copy(self):
        return self.__class__(self)

    def has_transparency(self):

        return self.a < 255


class MarginType:

    def __init__(self, margin):

        if margin is None:
            margin = 0, 0, 0, 0
        elif isinstance(margin, int):
            margin = tuple([margin] * 4)
        elif isinstance(margin, MarginType):
            margin = margin.left, margin.top, margin.right, margin.bottom
        else:
            l = len(margin)
            if l == 2:
                margin = tuple(margin) + tuple(margin)
            elif l == 3:
                margin = tuple(margin) + tuple([margin[1]])
            else:
                assert l == 4, f"Wrong value for margin type : {margin}"

        self.left = margin[0]
        self.top = margin[1]
        self.right = margin[2]
        self.bottom = margin[3]

    def __repr__(self):
        return f"MarginType(left={self.left}, top={self.top}, right={self.right}, bottom={self.bottom})"

    is_null = property(lambda self: self.left == self.top == self.right == self.bottom == 0)


def decorator_enable(widget, enable):
    functools.wraps(enable)
    def wrapped_func(*args, **kwargs):
        if widget.is_enabled: return
        widget._is_enabled = True
        res = enable(*args, **kwargs)
        widget.signal.ENABLE.emit()
        return res
    return wrapped_func


def decorator_disable(widget, disable):
    functools.wraps(disable)
    def wrapped_func(*args, **kwargs):
        if widget.is_enabled is False: return
        widget._is_enabled = False
        res = disable(*args, **kwargs)
        widget.signal.DISABLE.emit()
        return res
    return wrapped_func


class Enablable:

    def __init__(self):

        self.enable = decorator_enable(self, self.enable)
        self.disable = decorator_disable(self, self.disable)

        self._is_enabled = True
        self.create_signal("ENABLE")
        self.create_signal("DISABLE")

    is_enabled = property(lambda self: self._is_enabled)

    def enable(self):
        """Stuff to do when enable is called (decorated method)"""

    @staticmethod
    def enabled_from(enables_list):

        return tuple(Enablable.ienabled_from(enables_list))

    def disable(self):
        """Stuff to do when disable is called (decorated method)"""

    @staticmethod
    def ienabled_from(enables_list):

        for enablable in enables_list:
            if enablable.is_enabled:
                yield enablable


def decorator_validate(widget, validate):
    functools.wraps(validate)
    def wrapped_func(*args, **kwargs):
        if widget._catch_errors:
            res = None
            try:
                res = validate(*args, **kwargs)
            except Exception as e:
                if isinstance(e, ApplicationExit):
                    raise e
                LOGGER.warning("Error while executing {} validation: {}".format(widget, e))
        else:
            res = validate(*args, **kwargs)
        widget.signal.VALIDATE.emit(res)
        return res
    return wrapped_func


class Validable:

    def __init__(self, catching_errors=False):

        self.validate = decorator_validate(self, self.validate)

        self._catch_errors = catching_errors
        self.create_signal("VALIDATE")

        if isinstance(self, Focusable):
            self.connect("_check_validate", self.signal.KEYDOWN)

    catching_erros = property(lambda self: self._catch_errors)

    def _check_validate(self, key):
        """Checks if the keydown asks for validation"""
        if key is keyboard.RETURN:
            # Enter or validation key
            self.validate()

    def validate(self):  # TODO : handle_validate
        """Stuff to do when validate is called (decorated method)"""


class Focusable(Enablable):
    """
    A Focusable is a Widget who listen to keyboard input, when it has the focus. Only one
    widget can be focused in the same time. When a new clic occurs and it doesn't collide
    this widget, it is defocused. If disabled, it cannot be focused.

    It has no 'focus' method since it is the application who decide who is focused or not.
    However, it can defocus itself or send a request for focusing the next focusable in the
    container parent

    When the mouse click on a Text inside a Button inside a focusable Zone inside a Scene,
    then the only the youngest Focusable is focused
    Here, it is the Button -> display.focused_comp = Button
    """
    def __init__(self):

        Enablable.__init__(self)

        self._is_focused = False
        self.create_signal("FOCUS")
        self.create_signal("DEFOCUS")
        self.create_signal("KEYDOWN")
        self.create_signal("KEYUP")

        self.connect("defocus", self.signal.DISABLE)
        self.connect("handle_focus", self.signal.FOCUS)
        self.connect("handle_defocus", self.signal.DEFOCUS)
        self.connect("handle_keydown", self.signal.KEYDOWN)
        self.connect("handle_keyup", self.signal.KEYUP)

    is_focused = property(lambda self: self._is_focused)

    def defocus(self):
        """Send a request for defousing this widget"""
        if not self.is_focused:
            return
        self.scene._focus(None)

    def focus_antecedant(self):
        """
        Give the focus to the previous focusable (ranked by position) is self.parent
        """
        all_focs = self.parent.children.focusables
        all_focs.sort()
        all_focs = Enablable.enabled_from(self.ivisibles_from(all_focs))
        for foc in all_focs:
            assert foc.is_visible and foc.is_enabled
        if len(all_focs) > 1:
            self.scene._focus(all_focs[(all_focs.index(self) -1) % len(all_focs)])

    def focus_next(self):
        """
        Give the focus to the next focusable (ranked by position) is self.parent
        """
        all_focs = self.parent.children.focusables
        all_focs.sort()
        all_focs = Enablable.enabled_from(self.ivisibles_from(all_focs))
        for foc in all_focs:
            assert foc.is_visible and foc.is_enabled
        if len(all_focs) > 1:
            self.scene._focus(
                all_focs[(all_focs.index(self) + (1 if keyboard.mod.maj == 0 else -1))
                         % len(all_focs)])

    def handle_defocus(self):
        """Stuff to do when the widget loose the focus"""

    def handle_focus(self):
        """Stuff to do when the widget receive the focus"""

    def handle_keydown(self, key):
        """Stuff to do when a key is pressed"""
        # Default code :
        # Checks if the keydown asks for swap focus
        # TODO : should we say that the TAB has to focus the next (other method connected to KEYDOWN) or let an override ?
        if key == keyboard.TAB:
            self.focus_next()

    def handle_keyup(self, key):
        """Stuff to do when a key is released"""


class Linkable(Focusable):
    """
    Abstract class for components who need to capture mouse clicks

    A Linkable is linked when a mouse LEFT BUTTON DOWN collide with it,
    and unlinked when the LEFT BUTTON UP occurs

    It has no 'link' or 'link_motion' method since it is the mouse who manages links.
    However, it can unlink itself

    WARNING : If a Linkable parent contains a Linkable child, and the LEFT BUTTON
              DOWN has occured on the child, then only the child will be linked

    NOTE : when a component is linked, you can access it via mouse.linked_comp
    """

    def __init__(self):

        Focusable.__init__(self)

        self.is_linked = False
        self.create_signal("LINK")
        self.create_signal("LINK_MOTION")
        self.create_signal("UNLINK")

        self.connect("handle_link", self.signal.LINK)
        self.connect("handle_link_motion", self.signal.LINK_MOTION)
        self.connect("handle_unlink", self.signal.UNLINK)

    def handle_link(self):
        """Stuff to do when the widget gets linked"""

    def handle_link_motion(self, link_motion_event):
        """Stuff to do when the widget'link has changed"""

    def handle_unlink(self):
        """Stuff to do when the widget's link is over"""

    def unlink(self):
        """Send a request for unlinking this widget"""
        if self is not mouse.linked_comp:
            raise PermissionError(f"{self} is not linked")
        mouse._unlink()


class Clickable(Linkable, Validable):
    """
    Abstract class for components who need to handle when they are clicked

    A component is clicked when the link is released while the mouse is still
    hovering it (for more explanations about links, see Linkable)

    WARNING : If a Clickable parent contains a Clickable child, and the click
    has occured on the child, then only the child will handle the click
    """

    def __init__(self, catching_errors=False):

        Linkable.__init__(self)
        Validable.__init__(self, catching_errors=catching_errors)
        def unlink():
            if self.collidemouse(): self.validate()
        self.signal.UNLINK.connect(unlink)


class Closable:
    """
    A Closable is a widget whose 'close' function is called when its scene is closed
    """
    def close(self):
        """Stuff to do when the widget'scene is closed"""


class Dragable(Linkable):
    """
    Abstract class for components who want to be moved by mouse

    NOTE : a link_motion pointing a dragable's child will drag the parent if the child is not Linkable
    """
    # TODO : solve : a component can be set dragable two times, wich create an overdrag

    def __init__(self):

        Linkable.__init__(self)

        def drag(drag_event):
            self.move(*drag_event.rel)
        self.signal.LINK_MOTION.connect(drag)

    @staticmethod
    def set_dragable(widget):
        if not hasattr(widget, "is_linked"):
            Linkable.__init__(widget)
            widget.is_linked = False
        # widget.handle_link.add(lambda: setattr(widget, "is_linked", True))
        # widget.handle_unlink.add(lambda: setattr(widget, "is_linked", False))

        def drag(drag_event):
            widget.move(*drag_event.rel)
        widget.signal.LINK_MOTION.connect(drag)


class Openable:
    """
    An Openable is a widget whose 'open' function is called when its scene gets open
    """
    # TODO : rework
    def open(self):
        """Stuff to do when the widget's scene gets open"""

